from __future__ import annotations

import dataclasses
import inspect
from dataclasses import field
from typing import TypeVar, get_origin, Type


class InjectorError(Exception):
    pass


T = TypeVar('T')


class Injector:
    def __init__(self):
        self._registry = {}

    def bind(self, instance, *, to=None, named=None):
        """
        Register an instance with the injector.

        Args:
            instance: The instance to register
            to: Optional class to bind the instance to. If None, uses instance's class.
            named: Optional string key for named binding.

        Raises:
            InjectorError: If a dependency is already registered for the target class and named
        """
        if to is None:
            to = instance.__class__

        key = (to, named)
        if key in self._registry:
            raise InjectorError(f"Dependency already registered for {to!r} named={named!r}")

        self._registry[key] = instance

    def _unbind(self, bind, named=None):
        """
        Unregister a dependency.

        Args:
            bind: The class to unregister
            named: Optional string key for named binding.
        """
        key = (bind, named)
        if key in self._registry:
            del self._registry[key]
        else:
            raise InjectorError(f"No dependency registered for {bind!r} named={named!r}")

    def get(self, cls: type[T], named=None) -> T:
        """
        Get a registered instance for the given class (and optional named).

        Args:
            cls: The class to look up
            named: Optional string key for named binding.

        Returns:
            The registered instance

        Raises:
            InjectorError: If no dependency is registered for the class/named
        """
        # allow either a real class or a parametrized generic alias (e.g. SomeClass[Pet])
        if not (inspect.isclass(cls) or get_origin(cls) is not None):
            raise InjectorError(f"Expected a class, got `{cls}` of type {type(cls)}")

        key = (cls, named)
        if key not in self._registry:
            raise InjectorError(f"No dependency registered for {cls.__name__} named={named}")

        return self._registry[key]

    def _clear(self):
        """Clear all registered dependencies."""
        self._registry.clear()

    @property
    def field(self) -> Type[inject]:  # todo test
        return inject


default_injector = Injector()
injector = default_injector


class inject:
    """
    A data descriptor for injecting dependencies.

    Examples:
        class Service:
            # Inject a dependency using type annotation
            repository: Repository = inject()

        # Named injection
        class Client:
            api: ApiService = inject(name="prod")
    """

    def __init__(self, *, named=None):
        self.named = named
        self.injector = default_injector
        self.name = None
        self.cls = None

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner):
        if instance is None:  # and is_dataclass(owner): # it is not yet a dataclass
            is_dataclass_in_doing = hasattr(owner, dataclasses._PARAMS) and \
                                    not dataclasses.is_dataclass(owner)
            if is_dataclass_in_doing:
                return field(init=False, default=self)

        if self.cls is None:
            ann = inspect.get_annotations(owner, eval_str=True)
            name = self.name
            if name in ann:
                self.cls = ann[name]
            else:
                raise InjectorError(f"Cannot find type annotation for {name} in {owner.__name__}")

        return self.injector.get(self.cls, self.named)

    def __set__(self, instance, value):
        raise InjectorError(f"Cannot set {self.name} directly. Use the injector to register a new instance.")



from __future__ import annotations

import logging
from collections.abc import Callable

import js
from pyodide.ffi import create_proxy

from wwwpy.remote import eventlib

logger = logging.getLogger(__name__)


class DragFsm:
    IDLE = "IDLE"
    READY = "READY"
    DRAGGING = "DRAGGING"

    DRAG_THRESHOLD_PX = 5

    def __init__(self):
        self.state = self.IDLE
        self.drag_start = (-1, -1)

    def reset(self):
        self.state = self.IDLE
        self.drag_start = (-1, -1)

    def pointerdown_accepted(self, event):
        was = self.state
        if self.state == self.IDLE:
            # logger.debug(f"Pointer down on valid source")
            self.drag_start = (event.clientX, event.clientY)
            self.state = self.READY
            event.target.setPointerCapture(event.pointerId)
            event.preventDefault()
            # event.stopPropagation()
        logger.debug(f'pointerdown_accepted was={was} now={self.state}')

    def pointermove(self, event) -> str:
        was = self.state
        if self.state == self.READY:
            # Check if we've moved enough to consider this a drag
            dx = abs(event.clientX - self.drag_start[0])
            dy = abs(event.clientY - self.drag_start[1])

            if dx > self.DRAG_THRESHOLD_PX or dy > self.DRAG_THRESHOLD_PX:
                logger.debug("Drag threshold exceeded, entering DRAG_ACTIVE state")
                self.state = self.DRAGGING
        logger.debug(f'pointermove was={was} now={self.state}')
        return self.state

    def transitioned_to_dragging(self, event) -> bool:
        if self.state != self.READY:
            return False
        self.pointermove(event)
        return self.state == self.DRAGGING

    def pointerup(self, event):
        was = self.state
        self.reset()
        logger.debug(f'pointerup was={was} now={self.state}')


OnPointerDownAcceptType = Callable[[js.Event], bool]


class DragManager:
    """
    Manages pointer state and interactions (click or drag) between source and target elements.

    This class tracks the state of pointer interactions and emits events at appropriate times.
    It supports both click-based and drag-based interaction patterns.
    """

    def __init__(self, on_pointerdown_accept: OnPointerDownAcceptType | None = None):
        """Initialize the PointerManager with default state and callbacks."""
        self._fsm = DragFsm()

        # Event callbacks
        if on_pointerdown_accept is None:
            on_pointerdown_accept = lambda event: False  # Default rejects all

        self.on_pointerdown_accept: OnPointerDownAcceptType = on_pointerdown_accept

    def install(self):
        eventlib.add_event_listeners(self)

    def uninstall(self):
        eventlib.remove_event_listeners(self)

    @property
    def state(self):
        return self._fsm.state

    def _js_window__pointerdown(self, event):
        if self.on_pointerdown_accept(event):
            self._fsm.pointerdown_accepted(event)

    def _js_window__pointermove(self, event):
        """Handle pointer move events for dragging and hovering."""
        self._fsm.pointermove(event)

    def _js_window__pointerup(self, event):
        """Handle pointer up events to complete drag operations."""
        self._fsm.pointerup(event)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Callable, Optional

from wwwpy.common.collectionlib import ListMap


@dataclass(frozen=True)
class Help:
    description: str
    url: str


_empty_help = Help("", "")


@dataclass
class NameHelp:
    name: str
    help: Help = field(default=_empty_help)

    @property
    def python_name(self) -> str:
        return self.name.replace('-', '_')


@dataclass
class EventDef(NameHelp):
    """Definition of an event of an HTML element."""
    pass


@dataclass
class AttributeDef(NameHelp):
    """Definition of an attribute of an HTML element."""
    values: list[str] = field(default_factory=list)
    boolean: bool = False
    mandatory: bool = False
    default_value: Optional[str] = None


class NamedListMap(ListMap):
    def __init__(self, args):
        super().__init__(args, key_func=lambda x: x.name)


@dataclass
class ElementDefBase:
    tag_name: str
    python_type: str

    def new_html(self, data_name: str) -> str:
        return f"""<{self.tag_name} data-name="{data_name}"></{self.tag_name}>"""


@dataclass
class ElementDef(ElementDefBase):
    help: Help = field(default=_empty_help)
    gen_html: Optional[Callable[[ElementDef, str], str]] = None
    """A function that generates the HTML for the element. It takes the data-name of the element as argument."""

    attributes: NamedListMap[AttributeDef] = field(default_factory=list)
    events: NamedListMap[EventDef] = field(default_factory=list)

    def __post_init__(self):
        if not isinstance(self.attributes, ListMap):
            self.attributes = NamedListMap(self.attributes)
        if not isinstance(self.events, ListMap):
            self.events = NamedListMap(self.events)

    def new_html(self, data_name: str) -> str:
        gen_html = self.gen_html or ElementDef.default_gen_html
        return gen_html(self, data_name)

    @classmethod
    def default_gen_html(cls, element_def: ElementDef, data_name: str) -> str:
        return f'\n<{element_def.tag_name} data-name="{data_name}"></{element_def.tag_name}>'


class ElementLibrary:
    def __init__(self):
        self._elements: dict[str, ElementDef] = {}

    def _add(self, element: ElementDef) -> None:
        """Add an ElementDef into the internal dict by its tag_name."""
        self._elements[element.tag_name] = element

    @property
    def elements(self) -> tuple[ElementDef, ...]:
        return tuple(list(self._elements.values()))

    def by_tag_name(self, tag_name: str) -> Optional[ElementDef]:
        res = self._elements.get(tag_name, None)
        if res is None:
            from wwwpy.common.designer import el_common
            res = el_common._create_unknown_element_def(tag_name)
            self._add(res)
        return res


_element_library: ElementLibrary = None


def element_library() -> ElementLibrary:
    global _element_library
    if _element_library is None:
        _element_library = ElementLibrary()
        from .el_shoelace import _shoelace_elements_def
        from .el_standard import _standard_elements_def

        for loader in (_shoelace_elements_def, _standard_elements_def):
            for el in loader():
                _element_library._add(el)
    return _element_library



import logging

import js
import pytest

logger = logging.getLogger(__name__)


@pytest.fixture
def clean_document():
    clean_document_now('begin')
    yield None
    clean_document_now('end')


def clean_document_now(mode='mode/NA'):
    # logger.debug(f'_clean_document {mode}')
    js.document.documentElement.innerHTML = ''
    js.document.head.innerHTML = ''
    for attr in js.document.documentElement.attributes:
        js.document.documentElement.removeAttributeNode(attr)

    # logger.debug(f'body=```{js.document.body.innerHTML}```')



from __future__ import annotations

import logging
from dataclasses import dataclass, field

import js

from wwwpy.common.designer.html_edit import Position
from wwwpy.common.designer.locator_lib import Locator, Origin
from wwwpy.remote._elementlib import element_xy_center
from wwwpy.remote.component import Component, get_component
from wwwpy.remote.designer.locator_js import locator_from
from wwwpy.remote.jslib import is_instance_of, get_deepest_element

logger = logging.getLogger(__name__)


def default_position_resolver(width: float, height: float, x: float, y: float) -> Position:
    """the x and y coordinates are relative to the top-left corner of the element"""
    from wwwpy.common.designer.ui._drop_indicator_svg import position_for
    return position_for(width, height, x, y)


def find_first_user_component() -> Component | None:
    # todo should exclude wwwpy internal components
    for element in js.document.body.children:
        logger.debug(f'find_first_user_component: checking {_pretty(element)}')
        component = get_component(element)
        if component:
            return component

    return None


@dataclass
class LocatorEvent:
    locator: Locator
    """The origin must be Origin.source"""
    main_element: js.Element
    main_xy: tuple[float, float]
    """The xy coordinates of the main element, relative to the page"""
    secondary_elements: list[js.HTMLElement] = field(default_factory=list)
    position_resolver: callable = field(default=default_position_resolver)

    def __post_init__(self):
        assert self.locator is not None, 'locator must not be None'
        assert self.locator.origin == Origin.source
        assert is_instance_of(self.main_element,
                              js.HTMLElement), f'main_element must be an HTMLElement, got {type(self.main_element)}'

    def position(self) -> Position:
        rect = self.main_element.getBoundingClientRect()
        w, h = rect.width, rect.height
        ax, ay = self.main_xy
        x = ax - rect.left
        y = ay - rect.top
        return self.position_resolver(w, h, x, y)

    @staticmethod
    def from_pointer_event(js_event: js.PointerEvent) -> LocatorEvent | None:
        xy = js_event.clientX, js_event.clientY
        element = get_deepest_element(*xy)
        logger.debug(f'from_pointer_event: {js_event.type} at {xy}, element: {_pretty(element)}')
        if element == js.document.body or element == js.document.documentElement:
            component = find_first_user_component()
            if not component:
                logger.debug('from_pointer_event: no user component found, returning None')
            else:
                element = component.element
        if not element:
            return None
        l = LocatorEvent.from_element(element, xy)
        return l

    @staticmethod
    def from_element(element: js.Element, xy: tuple[float, float] | None = None) -> LocatorEvent | None:
        locator = locator_from(element)
        loc_origin = locator.rebase_to_origin()
        if not loc_origin:
            logger.warning(f'locator returned None for element: {_pretty(element)}')
            return None

        if xy is None:
            xy = element_xy_center(element)

        return LocatorEvent(loc_origin, element, xy)


def _pretty(node: js.Element):
    if hasattr(node, 'tagName'):
        identifier = node.dataset.name if node.hasAttribute('data-name') else node.id
        return f'{node.tagName.lower()}#{identifier}.{node.className}[{node.innerHTML.strip()[:20]}…]'
    return str(node)



from __future__ import annotations

from wwwpy.common.designer.ui.rect_readonly import RectReadOnly
from wwwpy.remote import component as wpc


class Floater(wpc.Component, auto_define=False):
    """This is a floating component that will adapt itself to the given situation"""

    def set_reference_geometry(self, rect: RectReadOnly):
        """Set the geometry of the element that we are attached to."""



from typing import overload, Union

from wwwpy.common.designer.ui.rect_readonly import RectReadOnly


class RectReadOnlyPy(RectReadOnly):
    @overload
    def __init__(self, x: float, y: float, width: float, height: float) -> None:
        ...

    @overload
    def __init__(self, rect: RectReadOnly) -> None:
        ...

    def __init__(
            self,
            x: Union[float, RectReadOnly],
            y: float = None,
            width: float = None,
            height: float = None
    ) -> None:
        # duck‐type the “rect” overload
        if hasattr(x, "x") and hasattr(x, "y") and hasattr(x, "width") and hasattr(x, "height"):
            rect = x  # type: RectReadOnly
            self._x = rect.x
            self._y = rect.y
            self._width = rect.width
            self._height = rect.height
        else:
            # x,y,width,height overload
            self._x = x  # type: ignore
            self._y = y  # type: ignore
            self._width = width  # type: ignore
            self._height = height  # type: ignore

    @property
    def x(self) -> float:
        return self._x

    @property
    def y(self) -> float:
        return self._y

    @property
    def width(self) -> float:
        return self._width

    @property
    def height(self) -> float:
        return self._height

    @property
    def left(self) -> float:
        return self.x

    @property
    def top(self) -> float:
        return self.y

    @property
    def right(self) -> float:
        return self.x + self.width

    @property
    def bottom(self) -> float:
        return self.y + self.height

    def toJSON(self) -> object:
        return {
            "x": self.x,
            "y": self.y,
            "width": self.width,
            "height": self.height,
        }

    @property
    def xy_center(self) -> tuple[float, float]:
        """Get the center x and y coordinates of the rect."""
        return self.x + self.width / 2, self.y + self.height / 2

    def __repr__(self) -> str:
        return f"RectReadOnlyPy(x={self.x}, y={self.y}, width={self.width}, height={self.height})"



from __future__ import annotations

import inspect
import logging
import types
import weakref
from dataclasses import dataclass
from functools import cached_property

logger = logging.getLogger(__name__)

import js
from pyodide.ffi import create_proxy
from pyodide.ffi.wrappers import add_event_listener, remove_event_listener, EVENT_LISTENERS


@dataclass
class Accept:
    target: js.EventTarget
    type: str


_js_attrs = dict()

BY_CONVENTION = object()


def convention_accept(name: str) -> Accept | None:
    """This accepts methods with the format _js_window__click, _js_document__keydown, etc."""
    # Check if the string starts with "_js_"
    if not name.startswith('_js_'):
        return None

    # Find the position of "__"
    double_underscore_pos = name.find('__', 4)  # Start searching after "_js_"
    if double_underscore_pos == -1:
        return None

    target_name = name[4:double_underscore_pos]
    event_type = name[double_underscore_pos + 2:]

    # Check if target_name and event_type are not empty
    if not target_name or not event_type:
        return None

    # Map target_name to JS object
    target_obj = getattr(js, target_name, None)
    if target_obj is None:
        return None

    target_obj_stored = _js_attrs.get(target_name, None)
    if target_obj_stored is None:
        _js_attrs[target_name] = target_obj
    else:
        if target_obj.js_id != target_obj_stored.js_id:
            msg = f'js_id mismatch for {target_name}: {target_obj.js_id} != {target_obj.js_id}'
            raise RuntimeError(msg)

    logger.debug(f'target_name={target_name} target_obj.js_id={target_obj.js_id} event_type={event_type}')
    return Accept(target_obj, event_type)


def _process_event_listeners(instance, action_func, accept=convention_accept):
    """
    Helper function to process event listeners for methods in target.

    Args:
        instance: The object containing methods to process
        action_func: Function to call with (target, event_type, method) for matched methods
        accept: A function that determines if a method matches the event handler pattern
    """

    for name in dir(instance):
        if name.startswith('__'):
            continue
        try:
            try:
                attr = inspect.getattr_static(instance, name)
            except AttributeError:
                continue
            if isinstance(attr, property):
                continue
            if isinstance(attr, types.FunctionType):
                bound_method = getattr(instance, name)
            else:
                continue

            accepted = accept(name)
            if accepted is not None or _has_handler_options(bound_method.__func__):
                if callable(bound_method):
                    logger.debug(f'calling {action_func} for {name}')
                    h = handler(bound_method)
                    try:
                        getattr(h, action_func)()
                    except KeyError as e:
                        logger.error(f'KeyError: {e} for {name}')
                        logger.info(f'target=`{instance.__class__.__name__}` has no event listener for {name}')
                    logger.info(f'EVENT_LISTENERS=`{EVENT_LISTENERS}`')
        except Exception as e:
            logger.exception(f'Error processing {name}: {e}')
            raise


def add_event_listeners(target, accept=convention_accept):
    """
    Add event listeners to methods in target that match the naming convention.

    Args:
        target: The object containing methods to be used as event handlers
        accept: A function that determines if a method should be used as an event handler
    """
    c = _counter(target)
    _counter(target, 1)

    if c > 0:
        logger.debug(f'target={target.__class__.__name__} already has {c} event listeners installed, skipping')
        return
    logger.debug(f'target={target.__class__.__name__} has no event listeners installed, adding')
    _process_event_listeners(target, 'install', accept)


def remove_event_listeners(target, accept=convention_accept):
    """
    Remove event listeners from methods in target that match the naming convention.

    Args:
        target: The object containing methods that were used as event handlers
        accept: A function that determines if a method was used as an event handler
    """
    c = _counter(target, -1)
    if c != 0:
        logger.debug(f'target={target.__class__.__name__} still has {c} event listeners installed, not removing')
        return
    logger.debug(f'target={target.__class__.__name__} has no event listeners installed, removing')
    _process_event_listeners(target, 'uninstall', accept)


def _counter(target, amount=0) -> int:
    if not hasattr(target, '_install_count'):
        target._install_count = 0
    target._install_count += amount
    return target._install_count


class HandlerOptions:
    def __init__(self, func: callable, target: js.EventTarget = BY_CONVENTION, type: str = BY_CONVENTION,
                 capture: bool = False):
        self.func = func  # assign so self._convention can be used
        if target is BY_CONVENTION: target = self._convention.target
        if type is BY_CONVENTION: type = self._convention.type

        self.target: js.EventTarget = target
        self.type: str = type
        self.capture = capture

    @cached_property
    def _convention(self) -> Accept:
        accept = convention_accept(self.func.__name__)
        if accept is None:
            raise ValueError(f"Unable to determine convention for for `{self.func.__name__}`")
        return accept

    def install_event(self, h: Handler):
        self.target.addEventListener(self.type, h.proxy)


def _get_handler_options(func) -> HandlerOptions:
    if not _has_handler_options(func):
        raise ValueError(f'handler_options not set for this function {func.__name__}')
    return func._handler_options


def _has_handler_options(func) -> bool:
    return hasattr(func, '_handler_options')


def handler_options(*, target=BY_CONVENTION, type=BY_CONVENTION, capture=False):
    """
    Decorator to set handler options for a function.

    Args:
        target: The DOM target to which the event listener will be attached.
        type: The type of event to listen for (e.g., 'click', 'keydown').
        capture: Whether to use capture phase for the event listener.
    """

    def decorator(func):
        if _has_handler_options(func):
            raise ValueError('handler_options already set for this function')
        func._handler_options = HandlerOptions(func, target, type, capture)
        return func

    return decorator


class Handler:
    _instance: weakref.ref
    _class_func: types.FunctionType

    def __init__(self, instance, class_func: callable):
        self._instance = weakref.ref(instance)
        self._class_func = class_func
        # Only bound methods: wrap in WeakMethod
        self._install_count = 0
        self._strong_ref = None
        self._proxy = None

    @property
    def bound_method(self) -> callable:
        inst = self._instance()
        if inst is None:
            raise ReferenceError(f"target instance has been garbage‑collected; func={self._class_func.__name__}")
        bm = self._class_func.__get__(inst, self._class_func.__class__)
        return bm

    def install(self):
        func_name = self._class_func.__name__
        if self._proxy is not None:
            raise RuntimeError(f"handler already installed for {func_name}")

        self._proxy = create_proxy(self.bound_method)  # this will create a circular reference, to avoid gc of func
        # self._execute_method(True)
        ho = self._ho()
        if ho.capture:
            ho.target.addEventListener(ho.type, self._proxy, ho.capture)
        else:
            ho.target.addEventListener(ho.type, self._proxy)

    def _execute_method(self, install: bool):
        method_name = 'addEventListener' if install else 'removeEventListener'
        ho = self._ho()
        m = ho.target.addEventListener if install else ho.target.removeEventListener
        if ho.capture:
            logger.debug(
                f'executing {method_name} for {self._class_func.__name__} target={ho.target.js_id} type={ho.type}, capture=YES')
            m(ho.type, self._proxy, ho.capture)
        else:
            logger.debug(
                f'executing {method_name} for {self._class_func.__name__} target={ho.target.js_id} type={ho.type}, capture=NO')
            m(ho.type, self._proxy)

    def uninstall(self):
        func_name = self._class_func.__name__
        if self._proxy is None:
            raise RuntimeError(f"handler not installed for {func_name}")
        # ho = self._ho()
        # ho.target.removeEventListener(ho.type, self._proxy, ho.capture)
        self._execute_method(False)
        self._proxy = None

    def _ho(self):
        if not _has_handler_options(self._class_func):
            ho = HandlerOptions(self._class_func)
        else:
            ho = _get_handler_options(self._class_func)
        return ho


from weakref import WeakKeyDictionary

_instances: WeakKeyDictionary[object, dict] = WeakKeyDictionary()


def _instances_dict(instance: object) -> dict:
    i = _instances.get(instance, None)
    if i is None:
        i = dict()
        _instances[instance] = i
    return i


def handler(bound_method: types.MethodType | callable) -> Handler:
    assert isinstance(bound_method, types.MethodType), "method_storage() requires a bound method"
    # only class instance methods (not class type methods here)
    instance = bound_method.__self__
    instance_dict = _instances_dict(instance)
    mf = bound_method.__func__
    handler = instance_dict.get(mf, None)
    if handler is None:
        handler = Handler(instance, mf)
        instance_dict[mf] = handler

    return handler



from __future__ import annotations

import dataclasses
import inspect
import logging
from enum import Enum
from functools import cached_property

import js

import wwwpy.remote.component as wpc
import wwwpy.remote.designer.ui.new_toolbox  # noqa
from wwwpy.common import modlib
from wwwpy.common.designer.comp_info import iter_comp_info_folder, CompInfo, LocatorNode, ComponentDef
from wwwpy.common.designer.element_library import ElementDefBase
from wwwpy.common.eventbus import EventBus
from wwwpy.common.injectorlib import injector
from wwwpy.remote import dict_to_js
from wwwpy.remote.component import get_component
from wwwpy.remote.designer.dev_mode_events import AfterDevModeShow
from wwwpy.remote.designer.ui.design_aware import DesignAware
from wwwpy.remote.designer.ui.intent import IntentEvent, Intent, IntentChangedEvent
from wwwpy.remote.designer.ui.intent_add_element import AddElementIntent
from wwwpy.remote.designer.ui.intent_manager import IntentManager
from wwwpy.remote.designer.ui.locator_event import LocatorEvent
from wwwpy.remote.jslib import get_deepest_element, closest_across_shadow

logger = logging.getLogger(__name__)


class HeaderClick(Enum):
    MARKER = 'MARKER'
    TEXT = 'TEXT'


class _DesignAware(DesignAware):

    def find_intent(self, hover_event: IntentEvent) -> Intent | None:
        target = hover_event.deep_target
        # def is_container(element: js.Element) -> bool:
        #     return element.tagName.lower() == AddUserComponentIntentUI.component_metadata.tag_name.lower()
        #
        # is_cont = is_contained(target, is_container)
        # if is_cont:
        #     logger.warning(f'find_intent: is_contained {target.tagName} {target.className}')
        # return

        if not target: return None
        # res = target.closest(AddUserComponentIntentUI.component_metadata.tag_name)
        res = closest_across_shadow(target, AddUserComponentIntentUI.component_metadata.tag_name)
        if not res: return None
        comp = get_component(res, AddUserComponentIntentUI)
        if not comp: return None
        return comp.intent

        # comp_tree_item: CompStructureItem = wpc.get_component(res)
        # x = hover_event.js_event.clientX - comp_tree_item._summary.getBoundingClientRect().left
        # if x < 20: return None
        # return comp_tree_item.add_intent

    # def is_selectable_js(self, js_event: js.PointerEvent) -> bool | None:
    #     w = _click_where(js_event)
    #     if w == HeaderClick.TEXT:
    #         return True
    #     return None
    def is_selectable_le(self, locator_event: LocatorEvent) -> bool | None:
        if locator_event.locator.match_component_type(AddUserComponentIntentUI):
            logger.warning(f'is_selectable_le: click on summary')
            return None
        struct_item = _get_comp_structure_item(locator_event.main_element)
        if struct_item:
            if locator_event.main_element == struct_item._summary:
                logger.warning(f'is_selectable_le: click on summary')
                return None
            l = locator_event.locator
            logger.warning(f'is_selectable_le: {l.tag_name} {l.class_name}')
            left = locator_event.main_element.getBoundingClientRect().left
            return locator_event.main_xy[0] - left > 20
        return None

    def locator_event_transformer(self, locator_event: LocatorEvent) -> LocatorEvent | None:
        if not isinstance(locator_event, LocatorEvent):
            raise TypeError(f'Expected LocatorEvent, got {type(locator_event)}')
        locator_node = _get_locator_node(locator_event)
        if locator_node:
            logger.warning(f'locator_event_transformer: {locator_node.locator}')
            new_locator_event = dataclasses.replace(locator_event, locator=locator_node.locator)
            return new_locator_event
        return None


def _get_locator_node(locator_event: LocatorEvent) -> LocatorNode | None:
    if locator_event.locator.match_component_type(CompStructureItem):
        if hasattr(locator_event.main_element, '_locator_node'):
            locator_node: LocatorNode = locator_event.main_element._locator_node
            return locator_node
    return None


def _set_locator_node(element: js.Element, locator_node: LocatorNode):
    element._locator_node = locator_node


# class _PointerEventInfo:
#     element: js.Element
#     comp_tree_item: CompStructureItem
def _get_comp_structure_item(contained_element: js.Element) -> CompStructureItem | None:
    res = contained_element.closest(CompStructureItem.component_metadata.tag_name)
    if not res: return None
    comp_tree_item: CompStructureItem = wpc.get_component(res, CompStructureItem)
    return comp_tree_item


def _click_where(js_event: js.PointerEvent) -> HeaderClick | None:
    target = get_deepest_element(js_event.clientX, js_event.clientY)
    if not target: return None
    comp_tree_item = _get_comp_structure_item(target)
    x = js_event.clientX - comp_tree_item._summary.getBoundingClientRect().left
    # if x < 20:
    #     return HeaderClick.MARKER
    # else:
    #     return HeaderClick.TEXT
    return HeaderClick.TEXT if x > 20 else HeaderClick.MARKER


_design_aware = _DesignAware()


class CompStructure(wpc.Component, tag_name='wwwpy-comp-structure'):
    _div: js.HTMLDivElement = wpc.element()
    _eventbus: EventBus = injector.field()
    span1: js.HTMLElement = wpc.element()

    def init_component(self):
        self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.element.shadowRoot.innerHTML = """
<style>
    
    details > details {
        margin-left: 1em;
    }
    
    .no-marker > summary {
      list-style: none;
      padding-left: 0.5em;
    }
    .no-marker > summary::-webkit-details-marker {
      display: none;
    }
</style>
<div>
<div data-name="span1">Create new Component</div>
<div data-name="_div"></div>
</div>
        """
        self._subscription = None

    def connectedCallback(self):
        self._subscription = self._eventbus.subscribe(self._scan_for_components, on=AfterDevModeShow)
        DesignAware.EP_REGISTRY.unregister(_design_aware)
        DesignAware.EP_REGISTRY.register(_design_aware)

    def disconnectedCallback(self):
        self._subscription.unsubscribe()
        self._subscription = None
        DesignAware.EP_REGISTRY.unregister(_design_aware)

    def _scan_for_components(self, _):
        logger.debug('scan_for_components')
        self._div.innerHTML = ''
        rem = modlib._find_package_directory('remote')
        if not rem:
            return
        for ci in iter_comp_info_folder(rem, 'remote'):
            cti = CompStructureItem()
            cti.set_comp_info(ci)
            self._div.appendChild(cti.element)

    async def span1__click(self, event):
        logger.debug(f'{inspect.currentframe().f_code.co_name} event fired %s', event)
        if js.window.confirm('Add new component file?\nIt will be added to your "remote" folder.'):
            from wwwpy.server.designer import rpc
            res = await rpc.add_new_component()
            js.window.alert(res)
    


class CompStructureItem(wpc.Component, tag_name='wwwpy-comp-structure-item'):
    comp_info: CompInfo
    _details: js.HTMLElement = wpc.element()
    _summary: js.HTMLElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """
<details data-name="_details">
    <summary data-name="_summary"></summary>
</details>
        """

    def set_comp_info(self, ci: CompInfo):
        self.comp_info = ci

        self._summary.innerText = ci.path.name + ' / ' + ci.class_name
        self._summary.appendChild(self.intent_ui.element)

        def rec(ln_list: list[LocatorNode], elem: js.Element, level):
            for child in ln_list:
                cst_node = child.cst_node
                data_name = cst_node.attributes.get('data-name', None) or ''
                dn = '' if not data_name else f' - {data_name} '
                content = cst_node.content or ''

                if content:
                    escape_table = str.maketrans({
                        '\r': ' ',
                        '\n': ' ',
                        '\t': ' ',
                    })
                    content = content.translate(escape_table)
                    max_content = 15
                    if len(content) > max_content:
                        content = content[:max_content] + '…'
                    content = f' / {content}'

                summary_text = f'{cst_node.tag_name}{dn}{content}'

                # language=html
                html = """<details open><summary></summary></details>"""
                ch = js.document.createRange().createContextualFragment(html)
                elem.appendChild(ch)
                details = elem.lastElementChild
                if len(cst_node.children) == 0:
                    details.classList.add('no-marker')
                else:
                    rec(child.children, details, level + 1)
                summary = details.firstElementChild
                _set_locator_node(summary, child)
                summary.innerText = summary_text

        try:
            rec(ci.locator_root.children, self._details, 1)
        except:
            logger.exception('Error in CompTreeItem.set_comp_info')
            self._details.innerText = 'Error in CompTreeItem.set_comp_info'

    @cached_property
    def add_intent(self) -> Intent:
        tag_name = self.comp_info.tag_name
        label = f'Add {tag_name}'
        intent = AddElementIntent(label)
        element_def_min: ElementDefBase = ComponentDef(tag_name, self.comp_info.class_full_name, self.comp_info)
        intent.element_def = element_def_min
        return intent

    @cached_property
    def intent_ui(self) -> AddUserComponentIntentUI:
        intent_ui = AddUserComponentIntentUI()
        intent_ui.intent = self.add_intent
        return intent_ui


class AddUserComponentIntentUI(wpc.Component):
    intent: Intent
    _intent_manager: IntentManager = injector.field()
    _span: js.Element = wpc.element()

    def init_component(self):
        self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.element.shadowRoot.innerHTML = """
        
<style>
 :host {
        --primary-color: #6366f1;
        --primary-hover: #818cf8;
        --secondary-color: #4f46e5;
        --border-color: #4b5563;
        --shadow-color: rgba(0, 0, 0, 0.3);
        --workspace-bg: #1e1e2e;
        --palette-bg: #27293d;
        --text-color: #e2e8f0;
        --item-bg: #2d3748;
        --item-hover-bg: #3a4358;
        --selected-bg: #4c1d95;
        --selected-border: #8b5cf6;
    }
    .intent-ui {
        align-items: center;
        justify-content: center;
        vertical-align: center;
        border: 1px solid var(--border-color);
        border-radius: 8px;
        cursor: pointer;
        transition: all 0.2s ease;
        user-select: none;
        background-color: var(--item-bg);
        color: var(--text-color);
        touch-action: none;
    }

    .intent-ui:hover {
       transform: translateY(-2px);
        box-shadow: 0 3px 5px var(--shadow-color);
        background-color: var(--item-hover-bg);
    }

    .intent-ui.selected {
        background-color: var(--selected-bg);
        border: 1px solid var(--selected-border);
        box-shadow: 0 0 0 2px var(--selected-border), 0 0 12px rgba(139, 92, 246, 0.5);
        position: relative;
        transform: scale(1.05);
    }
</style>
 <svg data-name="_span" class='intent-ui' xmlns="http://www.w3.org/2000/svg" width="20" height="20" viewBox="0 0 24 24"
      fill="none"
      stroke="currentColor" stroke-width="2" stroke-linecap="round"
      stroke-linejoin="round">
     <rect x="3" y="8" width="18" height="8" rx="2" ry="2"></rect>
     <line x1="12" y1="12" x2="12" y2="12"></line>
 </svg>
"""
        self.intent: Intent = None

    def connectedCallback(self):
        self._intent_manager.on(IntentChangedEvent).add(self._on_intent_changed_event)

    def disconnectedCallback(self):
        self._intent_manager.on(IntentChangedEvent).remove(self._on_intent_changed_event)

    def _on_intent_changed_event(self, event: IntentChangedEvent):
        if not self.intent:
            return
        old = self.selected
        self.selected = event.new == self.intent
        if old != self.selected:
            logger.warning(f'_on_intent_changed_event: {event} selected: {self.selected}')

    @property
    def selected(self) -> bool:
        return self._span.classList.contains('selected')

    @selected.setter
    def selected(self, value: bool):
        if value:
            self._span.classList.add('selected')
        else:
            self._span.classList.remove('selected')



from __future__ import annotations

import inspect
from typing import TypeVar, Generic

T = TypeVar('T')


class ExtensionPointRegistry(Generic[T]):

    def __init__(self, owner: type[T]):
        self._owner = owner
        self._extensions: list[T] = []
        self.base_type = owner

    def register(self, extension: T):
        if not isinstance(extension, self.base_type):
            raise ExtensionPointError(f'Expected {self.base_type}, got {type(extension)}')

        if extension in self._extensions:
            raise ExtensionPointError(f'Extension {extension} already registered')

        self._extensions.append(extension)

    def unregister(self, extension: T) -> bool:
        remove = extension in self._extensions
        if remove:
            self._extensions.remove(extension)
        return remove

    def _clear(self):
        """Used in tests to clear the registry"""
        self._extensions.clear()

    def __iter__(self):
        return iter(self._extensions)

    def __len__(self):
        return len(self._extensions)


class ExtensionPointError(Exception):
    pass


class ep_registry:

    def __init__(self):
        self._name = None
        self._ep_registry: ExtensionPointRegistry[...] = None

    def __set_name__(self, owner, name):
        self._name = name

    def __get__(self, instance, owner):
        if self._ep_registry is None:
            try:
                ann = inspect.get_annotations(owner, eval_str=True)
            except:
                raise ExtensionPointError(f'Locally defined class are not supported. Cannot introspect {owner}')
            self._class_decl_frame = None
            descr_type = ann.get(self._name, None)
            self._ep_registry = ExtensionPointRegistry(owner)
            if not descr_type is ExtensionPointRegistry[owner]:
                raise ExtensionPointError(f'Expected {ExtensionPointRegistry[owner]}, got {descr_type}')
        return self._ep_registry

    def __set__(self, instance, value):
        raise ExtensionPointError(f"Cannot set {self._name}")
#
# import types
# import sys
# import functools
# def get_annotations(obj, *, globals=None, locals=None, eval_str=False):
#     """Compute the annotations dict for an object.
#
#     obj may be a callable, class, or module.
#     Passing in an object of any other type raises TypeError.
#
#     Returns a dict.  get_annotations() returns a new dict every time
#     it's called; calling it twice on the same object will return two
#     different but equivalent dicts.
#
#     This function handles several details for you:
#
#       * If eval_str is true, values of type str will
#         be un-stringized using eval().  This is intended
#         for use with stringized annotations
#         ("from __future__ import annotations").
#       * If obj doesn't have an annotations dict, returns an
#         empty dict.  (Functions and methods always have an
#         annotations dict; classes, modules, and other types of
#         callables may not.)
#       * Ignores inherited annotations on classes.  If a class
#         doesn't have its own annotations dict, returns an empty dict.
#       * All accesses to object members and dict values are done
#         using getattr() and dict.get() for safety.
#       * Always, always, always returns a freshly-created dict.
#
#     eval_str controls whether or not values of type str are replaced
#     with the result of calling eval() on those values:
#
#       * If eval_str is true, eval() is called on values of type str.
#       * If eval_str is false (the default), values of type str are unchanged.
#
#     globals and locals are passed in to eval(); see the documentation
#     for eval() for more information.  If either globals or locals is
#     None, this function may replace that value with a context-specific
#     default, contingent on type(obj):
#
#       * If obj is a module, globals defaults to obj.__dict__.
#       * If obj is a class, globals defaults to
#         sys.modules[obj.__module__].__dict__ and locals
#         defaults to the obj class namespace.
#       * If obj is a callable, globals defaults to obj.__globals__,
#         although if obj is a wrapped function (using
#         functools.update_wrapper()) it is first unwrapped.
#     """
#     if isinstance(obj, type):
#         # class
#         obj_dict = getattr(obj, '__dict__', None)
#         if obj_dict and hasattr(obj_dict, 'get'):
#             ann = obj_dict.get('__annotations__', None)
#             if isinstance(ann, types.GetSetDescriptorType):
#                 ann = None
#         else:
#             ann = None
#
#         obj_globals = None
#         module_name = getattr(obj, '__module__', None)
#         if module_name:
#             module = sys.modules.get(module_name, None)
#             if module:
#                 obj_globals = getattr(module, '__dict__', None)
#         obj_locals = dict(vars(obj))
#         unwrap = obj
#     elif isinstance(obj, types.ModuleType):
#         # module
#         ann = getattr(obj, '__annotations__', None)
#         obj_globals = getattr(obj, '__dict__')
#         obj_locals = None
#         unwrap = None
#     elif callable(obj):
#         # this includes types.Function, types.BuiltinFunctionType,
#         # types.BuiltinMethodType, functools.partial, functools.singledispatch,
#         # "class funclike" from Lib/test/test_inspect... on and on it goes.
#         ann = getattr(obj, '__annotations__', None)
#         obj_globals = getattr(obj, '__globals__', None)
#         obj_locals = None
#         unwrap = obj
#     else:
#         raise TypeError(f"{obj!r} is not a module, class, or callable.")
#
#     if ann is None:
#         return {}
#
#     if not isinstance(ann, dict):
#         raise ValueError(f"{obj!r}.__annotations__ is neither a dict nor None")
#
#     if not ann:
#         return {}
#
#     if not eval_str:
#         return dict(ann)
#
#     if unwrap is not None:
#         while True:
#             if hasattr(unwrap, '__wrapped__'):
#                 unwrap = unwrap.__wrapped__
#                 continue
#             if isinstance(unwrap, functools.partial):
#                 unwrap = unwrap.func
#                 continue
#             break
#         if hasattr(unwrap, "__globals__"):
#             obj_globals = unwrap.__globals__
#
#     if globals is None:
#         globals = obj_globals
#     if locals is None:
#         locals = obj_locals
#
#     return_value = {key:
#                         value if not isinstance(value, str) else eval(value, globals, locals)
#                     for key, value in ann.items() }
#     return return_value



import xml.etree.ElementTree as ET
from typing import Callable, Dict


def set_width_height(svg: str, width: float, height: float) -> str:
    root = ET.fromstring(svg)
    root.set('width', str(width))
    root.set('height', str(height))
    # return ET.tostring(root, encoding='unicode')
    out = ET.tostring(root, encoding='unicode')
    out = out.replace('ns0:', '').replace('xmlns:ns0=', 'xmlns=')
    return out


def add_rounded_background2(svg: str, fill_color: str, ratio: float = 1.4, radius: float = 5) -> str:
    r = add_rounded_background3(svg, fill_color, ratio, radius)
    r2 = set_width_height(r, 22, 22)
    return r2


def add_rounded_background3(svg: str, fill_color: str, ratio: float = 1.4, radius: float = 5) -> str:
    root = ET.fromstring(svg)
    orig_w = float(root.attrib['width'])
    orig_h = float(root.attrib['height'])
    assert orig_w == orig_h
    if 'viewBox' in root.attrib:
        x0, y0, vb_w, vb_h = map(float, root.attrib['viewBox'].split())
    else:
        x0 = y0 = 0.0
        vb_w = vb_h = orig_w
    assert vb_w == vb_h == orig_w
    new_size = orig_w * ratio
    offset = (orig_w - new_size) / 2

    def fmt(v):
        return str(int(v)) if v.is_integer() else str(v)

    new_s = fmt(new_size)
    off_s = fmt(offset)
    root.attrib['viewBox'] = f'{off_s} {off_s} {new_s} {new_s}'
    root.attrib['width'] = new_s
    root.attrib['height'] = new_s
    ns = '{http://www.w3.org/2000/svg}'
    bg = ET.Element(ns + 'rect', {
        'x': off_s,
        'y': off_s,
        'width': new_s,
        'height': new_s,
        'fill': fill_color,
        'rx': str(radius),
        'ry': str(radius)
    })
    root.insert(0, bg)
    for e in root.iter():
        if isinstance(e.tag, str) and e.tag.startswith(ns):
            e.tag = e.tag[len(ns):]
    for k in list(root.attrib):
        if k.startswith('xmlns'):
            root.attrib.pop(k)
    root.attrib['xmlns'] = 'http://www.w3.org/2000/svg'
    return ET.tostring(root, encoding='unicode')


def add_rounded_background(svg: str, fill_color: str, radius: float) -> str:
    root = ET.fromstring(svg)
    x, y, w, h = root.attrib.get('viewBox', '0 0 0 0').split()
    ns = '{http://www.w3.org/2000/svg}'
    bg = ET.Element(ns + 'rect', {
        'x': x,
        'y': y,
        'width': w,
        'height': h,
        'fill': fill_color,
        'rx': str(radius),
        'ry': str(radius)
    })
    root.insert(0, bg)
    for e in root.iter():
        if isinstance(e.tag, str) and e.tag.startswith('{'):
            e.tag = e.tag.split('}', 1)[1]
    for k in list(root.attrib):
        if k.startswith('xmlns'):
            root.attrib.pop(k)
    root.attrib['xmlns'] = 'http://www.w3.org/2000/svg'
    return ET.tostring(root, encoding='unicode')


AttributeMutator = Callable[[Dict[str, str]], None]


def build_document(svg: str, attribute_mutator: AttributeMutator) -> str:
    root = ET.fromstring(svg)
    for elem in root.iter():
        attribute_mutator(elem.attrib)
    out = ET.tostring(root, encoding='unicode')
    out = out.replace('ns0:', '').replace('xmlns:ns0=', 'xmlns=')
    return out



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Tuple, Iterable


class RaiseOnAny:
    def __init__(self, payload: str | Exception | None = None, config: Config | None = None):
        if config:
            self._raise_on_any_config = config
        else:
            if payload is None:
                raise ValueError('message cannot be None')
            self._raise_on_any_config = Config(payload)

    def __getattribute__(self, name):
        return roa_get_config(self).process(name)

    def __repr__(self):
        return f'RaiseOnAny({roa_get_config(self).nested_message()})'


@dataclass
class Config:
    payload: str | Exception | None = None
    parent: Config | None = None
    sub_attr: str = 'self'
    accept_set: set[str] = field(default_factory=set)

    def accept(self, *name: str):
        for n in name:
            self.accept_set.add(n)

    def nested(self, name: str) -> RaiseOnAny | None:
        if name in self.accept_set:
            return RaiseOnAny(config=Config(None, self, name))
        return None

    def process(self, name):
        n = self.nested(name)
        if n:
            return n
        msg, root_config = self.nested_message(name)

        if isinstance(root_config.payload, Exception):
            raise root_config.payload

        raise Exception(msg)

    def nested_message(self, name: str | None = None) -> Tuple[str, Config]:
        attrs = []
        c = self
        while True:
            attrs.insert(0, c.sub_attr)
            if c.parent:
                c = c.parent
            else:
                break

        if name:
            attrs.append(name)
        name_fqn = '.'.join(attrs)
        msg = f'{c.payload} - Failures on `{name_fqn}` attribute'
        return msg, c


def roa_get_config(i: RaiseOnAny) -> Config:
    return object.__getattribute__(i, '_raise_on_any_config')


# raise_on_use

def raise_on_use(except_on: Iterable[str] = None):
    def decorator(func):
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                x = RaiseOnAny(e)
                if except_on:
                    roa_get_config(x).accept(*except_on)
                return x

        wrapper.__name__ = func.__name__
        return wrapper

    return decorator



from __future__ import annotations

import logging
import time

import js
from pyodide.ffi import create_proxy

import wwwpy.remote.component as wpc
from wwwpy.remote import dict_to_js, hotkeylib
from wwwpy.remote._elementlib import ensure_tag_instance
from wwwpy.remote.designer.ui.design_aware import DesignAware
from wwwpy.remote.designer.ui.intent import IntentEvent
from wwwpy.remote.designer.ui.locator_event import LocatorEvent
from wwwpy.remote.jslib import is_instance_of, is_contained

logger = logging.getLogger(__name__)


class _SidebarDesignAware(DesignAware):

    def is_selectable_le(self, locator_event: LocatorEvent) -> bool | None:
        def is_container(element: js.Element) -> bool:
            return element.tagName.lower() == PushableSidebar.component_metadata.tag_name.lower()

        is_cont = is_contained(locator_event.main_element, is_container)
        logger.debug(f'is_selectable_le is_contained inside :{is_cont} ')
        if is_cont:
            return False
        return None

    def is_selectable_old(self, hover_event: IntentEvent) -> bool | None:
        # return None
        target = hover_event.deep_target
        if target is None:
            return None

        sidebar = is_inside_sidebar(target)
        if sidebar:
            logger.debug(f'inside sidebar')
            return False
        else:
            logger.debug('outside sidebar')
            return None


_sidebar_design_aware = _SidebarDesignAware()


def register_extension_point():
    DesignAware.EP_REGISTRY.register(_sidebar_design_aware)


_BODY_PADDING_STYLE_ID = '_wwwpy_body_padding_style'


def is_inside_sidebar(element: js.Element) -> bool:
    tag = PushableSidebar.component_metadata.tag_name

    def _check(el: js.Element) -> bool:
        # light DOM
        if el.closest(tag):
            return True
        # shadow DOM: climb to host if present
        root = el.getRootNode()
        host = getattr(root, 'host', None)
        if host:
            return _check(host)
        return False

    return _check(element)


class PushableSidebar(wpc.Component, tag_name='pushable-sidebar'):
    position: str = wpc.attribute()
    width: str = wpc.attribute()
    min_width: str = wpc.attribute()
    max_width: str = wpc.attribute()
    collapsed_width: str = wpc.attribute()
    z_index: str = wpc.attribute()
    state: str = wpc.attribute()

    _toggle_button: js.HTMLButtonElement = wpc.element()
    _close_button: js.HTMLButtonElement = wpc.element()
    _resize_handle: js.HTMLDivElement = wpc.element()
    _container: js.HTMLDivElement = wpc.element()
    _sidebar_content: js.HTMLDivElement = wpc.element()
    _style: js.HTMLStyleElement = wpc.element()
    _sb_buttons: js.HTMLDivElement = wpc.element()

    def init_component(self):
        """

        # Update the style
        self._update_style()Initialize the sidebar component"""
        # Create shadow DOM with HTML template
        self.element.attachShadow(dict_to_js({'mode': 'open'}))

        # language=html
        self.element.shadowRoot.innerHTML = """
        <style data-name="_style"></style>
        
        <div class="sidebar-container" data-name="_container">
            <div class="sidebar-header">
                <div class="sidebar-header-buttons" data-name="_sb_buttons">
                    <button class="toggle-button" 
                            data-name="_toggle_button" 
                            title="Toggle sidebar">
                    </button>
                    <button class="close-button" 
                            data-name="_close_button" 
                            title="Hide sidebar">
                        &times;
                    </button>
                </div>
            </div>
            <div class="sidebar-content" data-name="_sidebar_content">
                <slot></slot>
            </div>
            <div class="resize-handle" data-name="_resize_handle"></div>
        </div>
        """

        self._config = {
            'width': '300px',
            'minWidth': '50px',
            'maxWidth': '5000px',
            'collapsedWidth': '30px',
            'zIndex': 9999,
            'enable_animation': True,
        }

        # Initial state
        self._state = 'expanded'

        # Resize state variables
        self._is_resizing = False
        self._start_width = 0
        self._start_x = 0

        # Apply attribute values if provided
        if self.width:
            self._config['width'] = self.width
        if self.min_width:
            self._config['minWidth'] = self.min_width
        if self.max_width:
            self._config['maxWidth'] = self.max_width
        if self.collapsed_width:
            self._config['collapsedWidth'] = self.collapsed_width
        if self.z_index:
            self._config['zIndex'] = self.z_index
        if self.state:
            self._state = self.state

        self._hotkeys = hotkeylib.Hotkey(js.window)
        self._last_ctrl_time = None
        self._hotkeys.add('CTRL-Control', self._double_ctrl_detector)

    @property
    def _body_padding(self) -> js.HTMLStyleElement:
        instance = ensure_tag_instance('style', _BODY_PADDING_STYLE_ID, js.document.head)
        assert is_instance_of(instance, js.HTMLStyleElement)
        return instance

    def _double_ctrl_detector(self, event):
        if not self._last_ctrl_time:
            self._last_ctrl_time = time.perf_counter()
            return
        _current_time = time.perf_counter()
        delta = _current_time - self._last_ctrl_time
        self._last_ctrl_time = _current_time
        if delta < 0.3:
            self._last_ctrl_time = None
            self.toggle()
            return

    @property
    def _position(self) -> str:
        return self.position or 'left'

    def _update_style(self):
        """Update the style element based on current configuration"""
        position = self._position
        animation_speed = 300 if self._config['enable_animation'] else 0

        # Generate CSS content
        # language=css
        css = f"""
         :host {{
            display: {'none' if self._state == 'hidden' else 'block'};
            position: fixed;
            top: 0;
            {position}: 0;
            height: 100%;
            box-sizing: border-box;
            z-index: {self._config['zIndex']};
            transition: width {animation_speed}ms ease;
            width: {self._config['collapsedWidth'] if self._state == 'collapsed' else self._config['width']};
            overflow: hidden;
        }}
        
        .sidebar-container {{
            display: flex;
            flex-direction: column;
            height: 100%;
            width: 100%;
            background-color: #333;
            color: #fff;
            box-shadow: {('2px' if position == 'left' else '-2px')} 0 5px rgba(0, 0, 0, 0.1);
            box-sizing: border-box;
            transition: transform {animation_speed}ms ease;
        }}
        
        .sidebar-header {{
            display: flex;
            justify-content: {('flex-end' if position == 'left' else 'flex-start')};
            align-items: center;
            padding: 5px;
            border-bottom: 1px solid #444;
        }}
        
        .sidebar-header-buttons {{
            display: flex;
            flex-direction: column;
            align-items: center;
        }}
        
        .toggle-button, .close-button {{
            background: none;
            border: none;
            cursor: pointer;
            color: #fff;
            font-size: 14px;
            padding: 3px;
            margin: 2px 0;
            width: 20px;
            height: 20px;
            display: flex;
            align-items: center;
            justify-content: center;
            border-radius: 2px;
        }}
        
        .toggle-button:hover, .close-button:hover {{
            background-color: rgba(255, 255, 255, 0.1);
        }}
        
        .close-button {{
            font-size: 16px;
        }}
        
        .sidebar-content {{
            flex: 1;
            overflow-y: auto;
            padding: 10px;
        }}
        
        .resize-handle {{
            position: absolute;
            top: 0;
            {('right' if position == 'left' else 'left')}: 0;
            width: 8px;
            height: 100%;
            cursor: {('e-resize' if position == 'left' else 'w-resize')};
            background-color: transparent;
            transition: background-color 0.2s;
            z-index: 2;
            touch-action: none;
        }}
        
        .resize-handle:hover,
        .resize-handle.active {{
            background-color: rgba(255, 255, 255, 0.2);
        }}
        
        :host([state="collapsed"]) .sidebar-content,
        :host([state="collapsed"]) .resize-handle {{
            opacity: 0;
        }}
        
        :host([state="collapsed"]) .sidebar-header {{
            border-bottom: none;
        }}
        """

        # Update the style element
        self._style.textContent = css

    def _update_sidebar(self):
        """Update sidebar appearance and behavior based on configuration"""
        # Set attribute to match internal state
        self.element.setAttribute('state', self._state)

        # Handle display based on state
        if self._state == 'hidden':
            self.element.style.display = 'none'
            self._remove_padding()
            return
        else:
            self.element.style.display = 'block'

        # Update the width based on state
        self.element.style.width = self._config['collapsedWidth'] if self._state == 'collapsed' else self._config[
            'width']

        # Update toggle button icon and title
        position = self._position
        toggle_icon = '&#9658;' if (position == 'left' and self._state == 'collapsed') or \
                                   (position == 'right' and self._state != 'collapsed') else '&#9668;'
        self._toggle_button.innerHTML = toggle_icon
        self._toggle_button.title = 'Expand sidebar' if self._state == 'collapsed' else 'Collapse sidebar'

        # Update style
        self._update_style()

        # Update document body padding
        self._adjust_content_padding()

    def _adjust_content_padding(self):
        """Adjust body padding to accommodate the sidebar"""
        # Remove existing padding first
        self._remove_padding()

        # Skip adding padding if sidebar is hidden
        if self._state == 'hidden':
            return

        # Add padding based on current sidebar state
        current_width = self._config['collapsedWidth'] if self._state == 'collapsed' else \
            self.element.style.width or self._config['width']

        # Calculate padding with the px suffix if needed
        padding_value = current_width if current_width.endswith('px') else f"{current_width}px"

        # Apply padding to body to push content
        self._set_padding(padding_value)

        # Dispatch event for external components to react
        self.element.dispatchEvent(
            js.CustomEvent.new('sidebar-resize', dict_to_js({
                'detail': {'width': current_width}
            }))
        )

    def _remove_padding(self):
        """Remove body padding"""
        self._set_padding('')

    async def _resize_handle__mousedown(self, event):
        """Start resize operation (using wwwpy's naming convention for event handlers)"""
        # If sidebar is in collapsed state, expand it first but keep the current width
        if self._state == 'collapsed':
            current_width = self.element.style.width
            self._config.update({'width': current_width})
            self.set_state('expanded')
            self.element.style.width = current_width  # Prevent jumping to stored width

        self._is_resizing = True
        self._start_width = int(js.getComputedStyle(self.element).width.replace('px', ''))
        self._start_x = event.clientX

        # Add active class to handle
        self._resize_handle.classList.add('active')

        # Set up mousemove and mouseup handlers directly on document
        self._mousemove_handler = create_proxy(self._handle_resize)
        self._mouseup_handler = create_proxy(self._stop_resize)

        js.document.addEventListener('mousemove', self._mousemove_handler)
        js.document.addEventListener('mouseup', self._mouseup_handler)
        js.document.addEventListener('mouseleave', self._mouseup_handler)

        # Prevent text selection during resize
        js.document.body.style.userSelect = 'none'

        # Add a resize class to the body to optimize rendering
        # self._config['animation_speed'] = 0  # Disable animation during resize
        # self._config['enable_animation'] = False
        self._config.update({'enable_animation': False})
        self._update_style()

        event.preventDefault()

    def _handle_resize(self, event):
        """Handle resize during mouse move"""
        if not self._is_resizing:
            return

        # Calculate new width
        if self.position == 'left':
            new_width = self._start_width + (event.clientX - self._start_x)
        else:
            new_width = self._start_width - (event.clientX - self._start_x)

        # Apply min/max constraints
        min_width = int(self._config['minWidth'].replace('px', ''))
        max_width = int(self._config['maxWidth'].replace('px', ''))

        if new_width < min_width:
            new_width = min_width
        if new_width > max_width:
            new_width = max_width

        # Update the sidebar width
        self.element.style.width = f"{new_width}px"

        # Update body padding
        self._set_padding(f"{new_width}px")

    def _set_padding(self, padding):
        # if self._config['position'] == 'left':
        #     js.document.body.style.paddingLeft = padding
        # else:
        #     js.document.body.style.paddingRight = padding
        # use a style tag to set the padding to body; use important
        if padding == '':
            style = ''
        else:
            style = (
                """body {\n padding-WHERE: PADDING !important; \n} """
                .replace('WHERE', self.position)
                .replace('PADDING', padding))

        self._body_padding.innerHTML = style

    def _stop_resize(self, event):
        """End resize operation"""
        if not self._is_resizing:
            return

        self._is_resizing = False
        self._resize_handle.classList.remove('active')

        # Clean up event listeners
        js.document.removeEventListener('mousemove', self._mousemove_handler)
        js.document.removeEventListener('mouseup', self._mouseup_handler)
        js.document.removeEventListener('mouseleave', self._mouseup_handler)

        # Restore text selection
        js.document.body.style.userSelect = ''

        # Remove resize class
        # self._config['animation_speed'] = 300  # Restore animation speed
        self._config.update({'enable_animation': True})
        self._update_style()

        # Store the final width in the config
        width_val = self.element.style.width
        if width_val.endswith('px'):
            width_px = int(width_val.replace('px', ''))
            self._config['width'] = f"{width_px}px"
        else:
            self._config['width'] = width_val

        # Save the current width to localStorage if desired
        try:
            js.localStorage.setItem('pushable-sidebar-width', self._config['width'])
        except:
            pass

        # Fire an event for the resize completion
        self.element.dispatchEvent(
            js.CustomEvent.new('sidebar-resize-end', dict_to_js({
                'detail': {'width': self._config['width']}
            }))
        )

    async def _toggle_button__click(self, event):
        """Handle toggle button click - uses wwwpy's automatic event binding"""
        self.toggle()

    async def _close_button__click(self, event):
        """Handle close button click - uses wwwpy's automatic event binding"""
        self.set_state('hidden')

    # Public API methods
    def set_state(self, state):
        """Set sidebar state ('hidden', 'collapsed', 'expanded')"""
        logger.debug(f'set_state called with state: {state}, current state: {self._state}')
        if state not in ['hidden', 'collapsed', 'expanded']:
            logger.warning('Invalid state. Valid states are: hidden, collapsed, expanded')
            return self

        old_state = self._state
        self._state = state
        self._update_sidebar()

        # Dispatch event
        self.element.dispatchEvent(
            js.CustomEvent.new('sidebar-state-change', dict_to_js({
                'detail': {
                    'oldState': old_state,
                    'newState': state
                }
            }))
        )

        return self

    def get_state(self):
        """Get current state"""
        return self._state

    def toggle(self):
        logger.debug(f'toggle called, current state: {self._state}')
        """Toggle sidebar state: expanded <-> collapsed, or hidden -> expanded"""
        if self._state == 'hidden':
            self.set_state('expanded')
        elif self._state == 'expanded':
            self.set_state('collapsed')
        else:  # collapsed
            self.set_state('expanded')
        return self

    def set_width(self, width):
        """Set sidebar width programmatically"""
        self._config['width'] = width
        if self._state == 'expanded':
            self.element.style.width = width
            self._adjust_content_padding()
        return self

    def set_position(self, position):
        """Set sidebar position"""
        if position not in ['left', 'right']:
            logger.warning('Sidebar position must be "left" or "right"')
            return self

        # Remove current positioning
        self._remove_padding()

        # Update position
        self.position = position
        # self.element.setAttribute('position', position)

        # Re-initialize the component
        # self._update_sidebar()

        return self

    # Lifecycle callbacks from the web components spec
    def connectedCallback(self):
        """Called when the element is added to the DOM"""
        # logger.warning(f'connectedCallback {self._state}')
        self._hotkeys.install()
        # Update the sidebar when connected
        self._update_sidebar()

        # Add resize event listener for responsive behavior
        self._window_resize_handler = create_proxy(lambda event: self._adjust_content_padding())
        js.window.addEventListener('resize', self._window_resize_handler)

    def disconnectedCallback(self):
        """Called when the element is removed from the DOM"""
        # logger.warning(f'disconnectedCallback {self._state}')
        self._hotkeys.uninstall()
        # Remove the padding from body when sidebar is removed
        self._remove_padding()

        # Clean up event listeners
        if hasattr(self, '_window_resize_handler'):
            js.window.removeEventListener('resize', self._window_resize_handler)

    def attributeChangedCallback(self, name, old_value, new_value):
        """Called when an observed attribute changes"""
        if old_value == new_value:
            return

        # Convert kebab-case to camelCase for some attribute names
        elif name == 'width':
            self._config['width'] = new_value
        elif name == 'min-width':
            self._config['minWidth'] = new_value
        elif name == 'max-width':
            self._config['maxWidth'] = new_value
        elif name == 'collapsed-width':
            self._config['collapsedWidth'] = new_value
        elif name == 'z-index':
            self._config['zIndex'] = new_value
        elif name == 'state':
            if new_value in ['hidden', 'collapsed', 'expanded']:
                self._state = new_value
            else:
                logger.warning('Invalid state. Valid values are: hidden, collapsed, expanded')

        # Update the sidebar if it's already connected
        if self.element.isConnected:
            self._update_sidebar()



# tree.py
# credits to: https://stackoverflow.com/a/59109706/316766

from pathlib import Path
from typing import Protocol, Iterable

# prefix components:
space = '    '
branch = '│   '
# pointers:
tee = '├── '
last = '└── '


class NodeProtocol(Protocol):
    def iterdir(self) -> Iterable['NodeProtocol']: ...

    def is_dir(self) -> bool: ...

    @property
    def name(self) -> str: ...


def tree(dir_path: NodeProtocol, prefix: str = '', file_size=False):
    """A recursive generator, given a directory Path object
    will yield a visual tree structure line by line
    with each line prefixed by the same characters
    """
    contents = list(dir_path.iterdir())
    # contents each get pointers that are ├── with a final └── :
    pointers = [tee] * (len(contents) - 1) + [last]
    for pointer, path in zip(pointers, contents):
        fs = '' if not file_size or path.is_dir() else f' ({path.stat().st_size})'
        yield prefix + pointer + path.name + fs
        try:
            if path.is_dir():  # extend the prefix and recurse:
                extension = branch if pointer == tee else space
                # i.e. space because last, └── , above so no more |
                yield from tree(path, prefix=prefix + extension, file_size=file_size)
        except Exception as e:
            yield prefix + f'Error: {e}'


def print_tree(path, printer=print, file_size=True):
    for line in tree(Path(path), file_size=file_size):
        printer(line)


def filesystem_tree_str(path: Path | str) -> str:
    path = Path(path)
    if path.exists():
        tree_str = '\n'.join(tree(path, prefix='    ', file_size=True))
        return f'Filesystem tree for {path}:\n{tree_str}'
    else:
        return f'Filesystem path do not exists `{path}`'


if __name__ == '__main__':
    print_tree(Path(__file__).parent.parent, file_size=True)



from __future__ import annotations

import inspect
import uuid
import weakref
from collections import defaultdict
from typing import Callable, Dict, Optional, Type, TypeVar, get_type_hints

T = TypeVar('T')


class Subscription:
    """Represents an active subscription that can be canceled."""

    def __init__(self, event_bus: EventBus, event_type: Type, callback_id: str):
        self._event_bus = weakref.ref(event_bus)
        self._event_type = event_type
        self._callback_id = callback_id
        self._active = True

    def unsubscribe(self) -> bool:
        """Cancels this subscription."""
        if not self._active:
            return False

        event_bus = self._event_bus()
        if event_bus is None:
            self._active = False
            return False

        result = event_bus._remove_subscription(self._event_type, self._callback_id)
        self._active = False
        return result

    @property
    def is_active(self) -> bool:
        return self._active and self._event_bus() is not None


class EventBus:
    """Type-based event bus that routes events based on their Python types."""

    def __init__(self):
        # Maps event types to {callback_id: callback}
        self._subscribers: Dict[Type, Dict[str, Callable]] = defaultdict(dict)

    def subscribe(self, callback: Optional[Callable] = None, *, on: Optional[Type] = None) -> Subscription:
        """
        Subscribe to events of a specific type.

        Args:
            callback: Function to call when matching events are published
            on: Type of events to subscribe to. If None, inferred from callback type hints.

        Returns:
            A Subscription object that can be used to unsubscribe.
        """
        # If used as a decorator
        if callback is None:
            def decorator(func):
                return self.subscribe(func, on=on)

            return decorator

        # Try to infer the event type from callback signature if not explicitly provided
        event_type = on
        if event_type is None:
            hints = get_type_hints(callback)
            # Look for the first parameter's type hint
            sig = inspect.signature(callback)
            params = list(sig.parameters.values())
            if params and params[0].name in hints:
                event_type = hints[params[0].name]

        if event_type is None:
            raise ValueError("Could not determine event type. Please provide 'on' parameter or use type hints.")

        callback_id = str(uuid.uuid4())
        self._subscribers[event_type][callback_id] = callback
        return Subscription(self, event_type, callback_id)

    def publish(self, event: any, *, on: Optional[Type] = None) -> int:
        """
        Publish an event to all subscribed handlers.

        Args:
            event: The event object to publish
            on: Optional type to publish as (must be compatible with event's type)

        Returns:
            int: Number of handlers notified
        """
        event_type = type(event) if on is None else on

        if on is not None and not isinstance(event, on):
            raise TypeError(f"Cannot publish {type(event)} as {on}")

        count = 0
        if event_type in self._subscribers:
            for callback in list(self._subscribers[event_type].values()):
                try:
                    callback(event)
                    count += 1
                except Exception as e:
                    # Log exception but continue with other handlers
                    import logging
                    logging.exception(f"Error in event handler {callback}: {e}")
                    count += 1  # Still count this as a notification attempt

        return count

    def _remove_subscription(self, event_type: Type, callback_id: str) -> bool:
        """Remove a subscription by its ID. Returns True if successful."""
        if event_type in self._subscribers and callback_id in self._subscribers[event_type]:
            del self._subscribers[event_type][callback_id]
            # Clean up empty dicts
            if not self._subscribers[event_type]:
                del self._subscribers[event_type]
            return True
        return False

    def clear(self) -> None:
        """Remove all subscriptions."""
        self._subscribers.clear()

    def get_subscriber_count(self, event_type: Type) -> int:
        """Get the number of subscribers for a specific event type."""
        return len(self._subscribers.get(event_type, {}))

    def connect(self, finalizable=None) -> EventBusConnection:
        return EventBusConnection(self, finalizable)


class EventBusConnection:
    def __init__(self, bus: EventBus, finalizable):
        self._bus = bus
        if finalizable is not None:  # register finalizer
            weakref.finalize(finalizable, self._disconnect)
        self._subscriptions = []

    # could return Subscription but need to add tests
    def subscribe(self, callback: Optional[Callable] = None, *, on: Optional[Type] = None) -> None:
        s = self._bus.subscribe(callback, on=on)
        self._subscriptions.append(s)

    def _disconnect(self):  # not yet public
        for s in self._subscriptions:
            s.unsubscribe()



import logging
from dataclasses import dataclass
from functools import cached_property

import js

from wwwpy.common.designer.canvas_selection import CanvasSelection
from wwwpy.common.injectorlib import injector
from wwwpy.remote.designer.ui.element_selector import ElementSelector
from wwwpy.remote.designer.ui.intent import Intent
from wwwpy.remote.designer.ui.locator_event import LocatorEvent

logger = logging.getLogger(__name__)


@dataclass
class SelectElementIntent(Intent):
    """Action to select an element in the designer."""
    label: str = 'Select'
    icon: str = 'select_element_icon'

    def on_hover(self, event: LocatorEvent):
        logger.debug(f'on_hover {event}')
        self._set_selection_from_js_event(event)

    def on_submit(self, event: LocatorEvent) -> bool:
        logger.debug(f'on_submit {event}')
        self._set_selection_from_js_event(event)
        injector.get(CanvasSelection).current_selection = event.locator
        return True

    def _set_selection_from_js_event(self, le: LocatorEvent):
        element_selector: ElementSelector = self._element_selector
        if not element_selector.element.isConnected:
            js.document.body.appendChild(element_selector.element)

        target = None
        if le is not None:
            target = le.main_element

        if element_selector.get_selected_element() != target:
            element_selector.set_selected_element(target)

        return target

    @cached_property
    def _element_selector(self) -> ElementSelector:
        return ElementSelector()


def _pretty(node: js.HTMLElement):
    if hasattr(node, 'tagName'):
        identifier = node.dataset.name if node.hasAttribute('data-name') else node.id
        return f'{node.tagName.lower()}#{identifier}.{node.className}[{node.innerHTML.strip()[:20]}…]'
    return str(node)



from __future__ import annotations

import logging

import js

from wwwpy.common.extension_point import ExtensionPointRegistry, ep_registry
from wwwpy.remote.designer.ui.intent import IntentEvent, Intent
from wwwpy.remote.designer.ui.locator_event import LocatorEvent

logger = logging.getLogger(__name__)


class DesignAware:
    EP_REGISTRY: ExtensionPointRegistry[DesignAware] = ep_registry()

    # def is_selectable_js(self, js_event: js.PointerEvent) -> bool | None:
    #     return None

    def is_selectable_le(self, locator_event: LocatorEvent) -> bool | None:
        return None

    def locator_event_transformer(self, locator_event: LocatorEvent) -> LocatorEvent | None:
        return None

    def find_intent(self, hover_event: IntentEvent) -> Intent | None:
        return None


def ep_is_selectable_le(locator_event: LocatorEvent) -> bool:
    except_none = [
        r for r in
        [(ep.__class__.__module__, ep.is_selectable_le(locator_event)) for ep in DesignAware.EP_REGISTRY]
        if r is not None
    ]
    bools = [r[1] for r in except_none]
    # if except_none and not any(except_none):  # list has at least one element and all are False
    #     return False
    if len(bools) == 0:
        return True
    if all(b is None for b in bools):
        logger.debug(f'all DesignAware.is_selectable_le returned None: {except_none}')
        return True

    if len(bools) > 0 and _all_false(bools):
        logger.debug(f'all DesignAware.is_selectable_le returned False or None: {except_none}')
        return False
    # logger.warning(f'all DesignAware.is_selectable returned True or None  : {except_none}')
    return True


def _all_false(lst: list[bool]) -> bool:
    """Check if all elements in the list are False."""
    return all(not item for item in lst)


def find_intent_da(js_event: js.PointerEvent) -> Intent | None:
    intent_event = IntentEvent(js_event)
    if intent_event.deep_target is None:
        return None
    for extension in DesignAware.EP_REGISTRY:
        try:
            intent = extension.find_intent(intent_event)
            if intent:
                return intent
        except:
            logger.exception('find_intent EP error', stacklevel=2)
    return None


def ep_locator_event_transformer(locator_event: LocatorEvent) -> LocatorEvent | None:
    for extension in DesignAware.EP_REGISTRY:
        try:
            new = extension.locator_event_transformer(locator_event)
            if new:
                return new
        except:
            logger.exception('find_intent EP error', stacklevel=2)
    return None



from __future__ import annotations

import logging

from wwwpy.common.type_listener import TypeListeners
from wwwpy.remote.designer.ui.design_aware import find_intent_da
from wwwpy.remote.designer.ui.intent import Intent, IntentChangedEvent
from wwwpy.remote.designer.ui.pointer_api import PointerApi, PointerDown, PointerMove, PointerUp

logger = logging.getLogger(__name__)


class IntentManager:

    def __init__(self) -> None:
        self._current_selection: Intent | None = None
        self._intent_changed_listeners = TypeListeners(IntentChangedEvent)
        self._ready_item: Intent | None = None

        self._pointer_api = PointerApi()
        self._pointer_api.on(PointerDown).add(self._on_pointer_down)
        self._pointer_api.on(PointerMove).add(self._on_pointer_move)
        self._pointer_api.on(PointerUp).add(self._on_pointer_up)

    def install(self) -> None:
        self._pointer_api.install()

    def uninstall(self) -> None:
        self._pointer_api.uninstall()

    @property
    def drag_state(self) -> str:
        return self._pointer_api.drag_state

    def on(self, event_type: type[IntentChangedEvent]) -> TypeListeners[IntentChangedEvent]:
        if event_type is not IntentChangedEvent:
            raise TypeError(f'IntentManager only supports {IntentChangedEvent}')
        return self._intent_changed_listeners

    def _toggle_selection(self, intent: Intent):
        if intent == self.current_selection:
            self.current_selection = None
        else:
            self.current_selection = intent

    def _on_pointer_down(self, event: PointerDown):
        intent = find_intent_da(event.js_event)
        logger.debug(f'_on_pointer_down {intent} state={self.drag_state}')
        if intent:
            self._ready_item = intent
            event.start_drag()
        else:
            se = self.current_selection
            if se is not None:
                if se.on_submit_js(event.js_event):
                    event.stop()
                    self.current_selection = None

    def _on_pointer_move(self, event: PointerMove):
        intent = find_intent_da(event.js_event)
        logger.debug(f'_on_pointer_move {intent} state={self.drag_state} '
                     f'ready_item={self._ready_item} drag_started={event.drag_started}')
        if event.drag_started and self._ready_item is not None:
            self.current_selection = self._ready_item
            self._ready_item = None

        if intent:
            return

        se = self.current_selection
        if se is not None:
            se.on_hover_js(event.js_event)

    def _on_pointer_up(self, event: PointerUp):
        intent = find_intent_da(event.js_event)
        logger.debug(f'_on_pointer_up {intent} state={self.drag_state} ready_item={self._ready_item}')

        if event.stopped: ...

        ready = self._ready_item
        self._ready_item = None

        if event.normal_click:
            if intent:
                if ready is not None:
                    self._toggle_selection(ready)
        elif event.drag_ended:
            if intent:
                return  # this return is not under test; when we pointerdown on an intent, and drag
                # (just enough) and release on the intent itself
            se = self.current_selection
            if se is not None:
                if se.on_submit_js(event.js_event):
                    self.current_selection = None

    @property
    def current_selection(self) -> Intent | None:
        return self._current_selection

    @current_selection.setter
    def current_selection(self, new: Intent | None) -> None:
        msg = ''
        if self._ready_item:
            msg += f' ri={self._ready_item}'

        old = self.current_selection
        if old == new:
            msg += f' (no change) old={old}'
            logger.debug(msg)
            return

        if old:
            old.selected = False
            old.on_deselected()
            msg += f' (deselecting {old})'

        self._current_selection = new
        msg += f' (selecting {None if new is None else new})'
        if new:
            new.selected = True
            new.on_selected()

        self._intent_changed_listeners.notify(IntentChangedEvent(old, new))

        logger.debug(msg)


def _pretty(node):
    if node is None:
        return 'None'
    if hasattr(node, 'tagName'):
        return f'{node.tagName.lower()}#{node.id}.{node.className}[{node.innerHTML.strip()[:20]}…]'
    return str(node)



from __future__ import annotations

import logging

import js
from pyodide.ffi import create_proxy

import wwwpy.remote.component as wpc
from wwwpy.remote import dict_to_js, dict_to_py
from wwwpy.remote.designer.ui.floater import Floater
from wwwpy.remote.designer.ui.floater_action_band import ActionBandFloater
from wwwpy.remote.designer.ui.floater_selection_indicator import SelectionIndicatorFloater
from wwwpy.remote.jslib import is_contained, AnimationFrameTracker

logger = logging.getLogger(__name__)


# logger.setLevel(logging.DEBUG)

# todo extract the logic of the animation frame tracker
class ElementSelector(wpc.Component, tag_name='element-selector'):
    selection_indicator: SelectionIndicatorFloater = wpc.element()
    action_band: ActionBandFloater = wpc.element()

    # _eventbus: EventBus = inject()

    def init_component(self):
        # Existing code remains the same
        self.element.attachShadow(dict_to_js({'mode': 'open'}))

        # language=html
        self.element.shadowRoot.innerHTML = """
        <selection-indicator-floater data-name="selection_indicator"></selection-indicator-floater>
        <action-band-floater data-name="action_band"></action-band-floater>
        """
        self._selected_element: js.HTMLElement | None = None
        self._last_position = None
        self._update_count = 0
        self._animation_frame_tracker = AnimationFrameTracker(self._on_animation_frame)

    def connectedCallback(self):
        has_py_comp = hasattr(self.element, '_python_component')
        logger.debug(f'has_py_comp: {has_py_comp}')

    def disconnectedCallback(self):
        self._animation_frame_tracker.stop()

    def is_selectable(self, element) -> bool:
        ok = not is_contained(element, self.element)
        return ok

    def set_selected_element(self, element):
        if element is not None and not self.is_selectable(element):
            raise ValueError(f'Element is not selectable `{dict_to_py(element)}`')

        if self._selected_element == element:
            return

        self._animation_frame_tracker.stop()
        self._selected_element = element
        self._last_position = None

        if element:
            self._animation_frame_tracker.start()
        else:
            self.selection_indicator.hide()
            self.action_band.hide()

    def get_selected_element(self):
        return self._selected_element

    def _on_animation_frame(self, timestamp):
        rect = self._selected_element.getBoundingClientRect()
        rect_tup = (rect.top, rect.left, rect.width, rect.height)
        if self._last_position == rect_tup:
            return

        skip_transition = self._last_position is not None

        self._update_count += 1
        logger.debug(f'update_highlight: {self._update_count} skip_transition: {skip_transition}')

        self.selection_indicator.transition = not skip_transition

        for x in [self.selection_indicator, self.action_band]:
            x: Floater
            x.set_reference_geometry(rect)

        self._last_position = rect_tup





import inspect
import logging

import js
from pyodide.ffi import create_proxy

import wwwpy.remote.component as wpc
from wwwpy.remote import dict_to_js

logger = logging.getLogger(__name__)


class LoggerLevelsComponent(wpc.Component):
    _ta_log: js.HTMLTextAreaElement = wpc.element()
    _title: js.HTMLDivElement = wpc.element()
    _row_container: js.HTMLDivElement = wpc.element()
    _search: js.HTMLInputElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """
<div data-name="_title">_title</div>
<input data-name="_search" placeholder="Filter by logger name">
<div data-name="_row_container"><br></div>  

<label>Component logs:</label>
<textarea data-name="_ta_log" placeholder="textarea1" rows="8" wrap="off" style="width: 98%"></textarea> 
"""
        self._list_all_logger()

    def _list_all_logger(self):
        import logging

        items = [(n, l) for n, l in logging.root.manager.loggerDict.items() if isinstance(l, logging.Logger)]
        # items.sort(key=lambda x: (x[1].level, x[0]), reverse=True)  # sort by logger name
        items.sort(key=lambda t: - t[1].level)
        items.sort(key=lambda t: t[0].lower())
        logger_counter = 0
        self._row_container.innerHTML = ''  # clear previous rows
        for name, logger in items:
            if not self._filter_accept(name):
                continue
            logger_counter += 1
            level_name = '' if logger.level == logging.NOTSET else logging.getLevelName(logger.level)
            row = LogConfRow()
            row.logger_name = name
            row.level_select = level_name
            self._row_container.appendChild(row.element)
        self._title.innerText = f'Found {logger_counter} loggers'
        self._log(self._title.innerText)

        # scroll to the top of the log area
        # self._ta_log.scrollTop = 0

    def _log(self, message):
        if not isinstance(message, str):
            message = str(message)
        self._ta_log.value += f"{message}\n"
        self._ta_log.scrollTop = self._ta_log.scrollHeight

    async def _row_container__input_row(self, event):
        row: LogConfRow = event.detail.row
        logging.getLogger(row.logger_name).setLevel(row.level_select)
        msg = f'{row.level_select:<8} {row.logger_name}'
        logger.debug(msg)
        self._log(msg)

    async def _search__input(self, event):
        logger.debug(f'{inspect.currentframe().f_code.co_name} event fired %s', event)
        self._list_all_logger()

    def _filter_accept(self, logger_name: str) -> bool:
        s = self._search.value.lower()
        if not s:
            return True
        ln = logger_name.lower()
        if ' ' in s:
            if s.strip() == ln:
                return True
            return False
        if s in ln:
            return True
        return False


class LogConfRow(wpc.Component):
    _level_select: js.HTMLSelectElement = wpc.element()
    _logger_name: js.HTMLSpanElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """       
<div>         
    <select data-name="_level_select"></select>
    <span data-name="_logger_name">_logger_name</span>
</div>
"""

        self._level_select.innerHTML = ''
        items = list(logging.getLevelNamesMapping().items())
        items.sort(key=lambda x: x[1])
        for name, level in items:
            option = js.document.createElement('option')
            option.value = name
            if level != logging.NOTSET:
                option.innerText = name
            self._level_select.appendChild(option)

    @property
    def level_select(self) -> str:
        return self._level_select.value

    @level_select.setter
    def level_select(self, level_name: str):
        self._level_select.value = level_name

    @property
    def logger_name(self) -> str:
        return self._logger_name.innerText

    @logger_name.setter
    def logger_name(self, name: str):
        self._logger_name.innerText = name

    async def _level_select__input(self, event):
        logger.debug(f'{inspect.currentframe().f_code.co_name} event fired %s', event)
        init = dict_to_js({'bubbles': True, 'detail': {'row': self}})
        self.element.dispatchEvent(js.CustomEvent.new('input-row', init))



import logging
from dataclasses import dataclass
from enum import Enum
from typing import TypeVar

import js

from wwwpy.common.type_listener import TypeListeners, DictListeners
from wwwpy.remote import eventlib
from wwwpy.remote.designer.ui.drag_manager import DragFsm
from wwwpy.remote.eventlib import handler_options

logger = logging.getLogger(__name__)


@dataclass
class PAEvent:
    js_event: js.PointerEvent


TPA = TypeVar('TPA', bound=PAEvent)


@dataclass
class PointerDown(PAEvent):
    _stop = False
    _start_drag = False

    def stop(self):
        if self._start_drag:
            raise RuntimeError('Cannot stop, start_drag already set')
        self._stop = True

    def start_drag(self):
        if self._stop:
            raise RuntimeError('Cannot start drag, stop already set')
        self._start_drag = True


@dataclass
class PointerMove(PAEvent):
    drag_started: bool


class Reason(str, Enum):
    normal_click = 'NORMAL_CLICK'
    drag_ended = 'DRAG_ENDED'
    stopped = 'STOPPED'


@dataclass
class PointerUp(PAEvent):
    reason: Reason

    @property
    def stopped(self) -> bool:
        return self.reason == Reason.stopped

    @property
    def drag_ended(self) -> bool:
        return self.reason == Reason.drag_ended

    @property
    def normal_click(self) -> bool:
        return self.reason == Reason.normal_click


class PointerApi:
    def __init__(self) -> None:
        self._listeners = DictListeners()
        self._drag_fsm = DragFsm()
        self._stopped = False
        self._stop_next_click = False

    def install(self) -> None:
        eventlib.add_event_listeners(self)

    def uninstall(self) -> None:
        eventlib.remove_event_listeners(self)

    @property
    def drag_state(self) -> str:
        return self._drag_fsm.state

    def on(self, event_type: type[TPA]) -> TypeListeners[TPA]:
        return self._listeners.on(event_type)

    def _notify(self, ev: PAEvent) -> None:
        self._listeners.notify(ev)

    @handler_options(capture=True)
    def _js_window__click(self, event: js.MouseEvent):
        logger.debug(
            f'_js_window__click _stop_next_click={self._stop_next_click} _stopped={self._stopped} state={self._drag_fsm.state}')
        if self._stop_next_click:
            _stop_js_event(event)
        self._stop_next_click = False

    @handler_options(capture=True)
    def _js_window__pointerdown(self, event: js.PointerEvent):
        e = PointerDown(event)
        self._notify(e)
        if e._start_drag:
            self._drag_fsm.pointerdown_accepted(event)
        elif e._stop:
            _stop_js_event(event)
            self._stopped = True
            self._stop_next_click = True

    def _js_window__pointermove(self, event: js.PointerEvent):
        dragging = self._drag_fsm.transitioned_to_dragging(event)
        e = PointerMove(event, dragging)
        self._notify(e)

    @handler_options(capture=True)
    def _js_window__pointerup(self, event: js.PointerEvent):
        r = None
        if self._drag_fsm.state == DragFsm.DRAGGING:
            r = Reason.drag_ended
        elif self._drag_fsm.state == DragFsm.READY:
            r = Reason.normal_click
        elif self._stopped:
            r = Reason.stopped
            _stop_js_event(event)
            self._stopped = False

        self._drag_fsm.pointerup(event)
        if r:
            self._notify(PointerUp(event, r))
        else:
            logger.debug(f'_js_window__pointerup: no event to notify fsm={self._drag_fsm.state}')


def _stop_js_event(e: js.Event):
    e.stopPropagation()
    e.preventDefault()
    e.stopImmediatePropagation()


def _pretty(node):
    if node is None:
        return 'None'
    if hasattr(node, 'tagName'):
        return f'{node.tagName.lower()}#{node.id}.{node.className}[{node.innerHTML.strip()[:20]}…]'
    return str(node)



import os

_INITIAL = set(globals())

PLAYWRIGHT_PATCH_TIMEOUT_MILLIS = 45000
PLAYWRIGHT_HEADFUL = False

_NEW = set(globals()) - _INITIAL
_NEW.remove('_INITIAL')
try:
    import wwwpy_user_conf

    # list all the local custom variables and copy them (if they exists from wwwpy_user_conf)
    for key in _NEW:
        if hasattr(wwwpy_user_conf, key):
            value = getattr(wwwpy_user_conf, key)
            globals()[key] = value
        else:
            env_value = os.environ.get(key)
            if env_value is not None:
                globals()[key] = env_value

except ModuleNotFoundError:
    import logging

    logger = logging.getLogger(__name__)
    logger.info("No local wwwpy_user_conf.py file found. Using default values.")



import dataclasses
import logging
from dataclasses import dataclass, field

import js

from wwwpy.common import modlib
from wwwpy.common.asynclib import create_task_safe
from wwwpy.common.designer.canvas_selection import CanvasSelection
from wwwpy.common.designer.code_edit import add_element, AddResult, AddFailed
from wwwpy.common.designer.element_library import ElementDefBase
from wwwpy.common.designer.html_edit import Position
from wwwpy.common.designer.html_locator import path_to_index
from wwwpy.common.designer.locator_lib import Locator
from wwwpy.common.injectorlib import injector
from wwwpy.remote.designer.ui.floater_drop_indicator import DropIndicatorFloater
from wwwpy.remote.designer.ui.intent import Intent
from wwwpy.remote.designer.ui.locator_event import LocatorEvent

logger = logging.getLogger(__name__)

_tool = DropIndicatorFloater()
_tool.transition = True


@dataclass
class AddElementIntent(Intent):  # todo rename InsertElementIntent
    element_def: ElementDefBase = None
    add_element: callable = field(default_factory=lambda: _add_component)

    def __post_init__(self):
        self._tool = _tool

    def on_selected(self):
        pass

    def on_deselected(self):
        self._tool.hide()

    def on_hover(self, locator_event: LocatorEvent):
        assert isinstance(locator_event, LocatorEvent), f'Expected LocatorEvent, got {type(locator_event)}'
        tool = self._tool
        if self._is_recursive(locator_event):
            logger.debug(f'Ignoring hover on {locator_event} because it is recursive')
            tool.hide()
            return
        if not tool.element.isConnected:
            js.document.body.appendChild(tool.element)
        tool.set_reference_geometry2(
            locator_event.main_element.getBoundingClientRect(),
            locator_event.position()
        )
        tool.show()

    def on_submit(self, locator_event: LocatorEvent) -> bool:
        if self._is_recursive(locator_event):
            logger.debug(f'Ignoring submit on {locator_event} because it is recursive')
            self._tool.hide()
            return False
        self.add_element(locator_event.locator, locator_event.position(), self.element_def)
        self._tool.hide()
        return True

    def _is_recursive(self, locator_event: LocatorEvent) -> bool:
        rec = locator_event.locator.class_full_name == self.element_def.python_type
        logger.debug(
            f'Recursive: {rec} {locator_event.locator.class_full_name} is recursive against {self.element_def.python_type}')
        return rec


def _pretty(node: js.HTMLElement):
    if hasattr(node, 'tagName'):
        identifier = node.dataset.name if node.hasAttribute('data-name') else node.id
        return f'{node.tagName.lower()}#{identifier}.{node.className}[{node.innerHTML.strip()[:20]}…]'
    return str(node)


def _add_component(el_path: Locator, position: Position,
                   element_def: ElementDefBase):
    file = modlib._find_module_path(el_path.class_module)
    old_source = file.read_text()

    path_index = path_to_index(el_path.path)
    add_result = add_element(old_source, el_path.class_name, element_def, path_index, position)

    if isinstance(add_result, AddResult):
        logger.debug(f'write_module_file len={len(add_result.source_code)} el_path={el_path}')
        injector.get(CanvasSelection).current_selection = dataclasses.replace(el_path, path=add_result.node_path)

        async def write_source_file():
            from wwwpy.server.designer import rpc
            write_res = await rpc.write_module_file(el_path.class_module, add_result.source_code)
            logger.debug(f'write_module_file res={write_res}')

        create_task_safe((write_source_file()))
    elif isinstance(add_result, AddFailed):
        js.alert('Sorry, an error occurred while adding the component.')



from __future__ import annotations

import logging

import js
from pyodide.ffi import create_proxy

import wwwpy.remote.component as wpc
from wwwpy.common.injectorlib import inject
from wwwpy.remote import dict_to_js
from wwwpy.remote.component import get_component
from wwwpy.remote.designer.ui.design_aware import DesignAware
from wwwpy.remote.designer.ui.intent import Intent, IntentChangedEvent, IntentEvent
from wwwpy.remote.designer.ui.intent_manager import IntentManager
from wwwpy.remote.jslib import get_deepest_element

logger = logging.getLogger(__name__)


class _PaletteIntentAware(DesignAware):

    def find_intent(self, hover_event: IntentEvent) -> Intent | None:
        return palette_find_indent(hover_event.deep_target)


def palette_find_indent(target: js.Element) -> Intent | None:
    res = target.closest(PaletteItemComponent.component_metadata.tag_name)
    if res:
        comp: PaletteItemComponent = get_component(res)
        return comp.intent
    return None


_palette_design_aware = _PaletteIntentAware()


def extension_point_register():
    DesignAware.EP_REGISTRY.register(_palette_design_aware)


def extension_point_unregister():
    DesignAware.EP_REGISTRY.unregister(_palette_design_aware)


class PaletteComponent(wpc.Component, tag_name='wwwpy-palette'):
    _item_container: js.HTMLDivElement = wpc.element()
    intent_manager: IntentManager = inject()

    def init_component(self):
        self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.element.shadowRoot.innerHTML = """
<div class="container">
    <div class="palette" data-name="_item_container">
    </div>
</div>
"""
        self.element.shadowRoot.innerHTML += _css_styles

        def ace(e: IntentChangedEvent):
            if e.old is not None:
                intent = self._action2item.get(id(e.old), None)
                if intent:
                    intent.selected = False
            if e.new is not None:
                intent = self._action2item.get(id(e.new), None)
                if intent:
                    intent.selected = True

        self._action2item = {}

        self._on_action_changed_event = ace
        self._palette_design_aware = _PaletteIntentAware()
        self._on_identify_event = lambda e: e.set_action(_find_palette_action(e.js_event))

    def connectedCallback(self):
        self.intent_manager.on(IntentChangedEvent).add(self._on_action_changed_event)

    def disconnectedCallback(self):
        self.intent_manager.on(IntentChangedEvent).remove(self._on_action_changed_event)

    def add_intent(self, intent: Intent) -> PaletteItemComponent:
        """Add an item to the palette."""
        item = PaletteItemComponent()
        item.intent = intent
        item.label = intent.label
        item.element.classList.add('palette-item')
        self._item_container.appendChild(item.element)
        self._action2item[id(intent)] = item
        # item.element.addEventListener('click', create_proxy(lambda e: self.intent_manager._action_item_click(item)))
        return item


class PaletteItemComponent(wpc.Component, tag_name='palette-item-icon'):
    _label: js.HTMLLabelElement = wpc.element()
    intent: Intent

    # override magic method so the f strings get a nice representation
    def __repr__(self):
        return f'{self.__class__.__name__}({self.label})'

    def init_component(self):
        # language=html
        self.element.innerHTML = """
 <svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" viewBox="0 0 24 24"
      fill="none"
      stroke="currentColor" stroke-width="2" stroke-linecap="round"
      stroke-linejoin="round">
     <rect x="3" y="8" width="18" height="8" rx="2" ry="2"></rect>
     <line x1="12" y1="12" x2="12" y2="12"></line>
 </svg>
 <label data-name="_label" style="display: block; text-align: center"></label>
 </div>
 </div> 
"""
        self.intent: Intent = None

    @property
    def label(self) -> str:
        return self._label.innerText

    @label.setter
    def label(self, value: str):
        self._label.innerText = value

    @property
    def selected(self) -> bool:
        return self.element.classList.contains('selected')

    @selected.setter
    def selected(self, value: bool):
        if value:
            self.element.classList.add('selected')
        else:
            self.element.classList.remove('selected')


# class ActionManager:
#     """A class to manage interaction and events to handle, drag & drop, element selection, move element."""
#
#     def __init__(self):
#         self.pointer_manager: PointerManager[Action] = PointerManager()
#         self.pointer_manager.on(IdentifyEvent).add(lambda e: e.set_action(_find_palette_action(e.js_event)))
#         self.install = self.pointer_manager.install
#         self.uninstall = self.pointer_manager.uninstall
#         self.on = self.pointer_manager.on
#
#     @property
#     def selected_action(self) -> Action | None:
#         return self.pointer_manager.selected_action
#
#     @selected_action.setter
#     def selected_action(self, value: Action | None):
#         self.selected_action = value


def _find_palette_action(event: js.Event) -> Intent | None:
    target = get_deepest_element(event.clientX, event.clientY)
    if target is None:  # tests missing. It looks like it happens when the mouse exit the viewport or moves on the scrollbar
        return None
    # logger.debug(f'_find_palette_item target={_pretty(target)}')
    import wwwpy.remote.designer.ui.palette as palette
    res = target.closest(palette.PaletteItemComponent.component_metadata.tag_name)
    if res:
        comp: PaletteItemComponent = get_component(res)
        return comp.intent
    return None


# language=html
_css_styles = """<style>
    :host {
        --primary-color: #6366f1;
        --primary-hover: #818cf8;
        --secondary-color: #4f46e5;
        --border-color: #4b5563;
        --shadow-color: rgba(0, 0, 0, 0.3);
        --workspace-bg: #1e1e2e;
        --palette-bg: #27293d;
        --text-color: #e2e8f0;
        --item-bg: #2d3748;
        --item-hover-bg: #3a4358;
        --selected-bg: #4c1d95;
        --selected-border: #8b5cf6;
    }

    body {
        font-family: 'Segoe UI', Tahoma, Geneva, Verdana, sans-serif;
        margin: 0;
        padding: 0;
        display: flex;
        flex-direction: column;
        height: 100vh;
        background-color: #111827;
        color: var(--text-color);
    }

    .header {
        background-color: var(--primary-color);
        color: white;
        padding: 1rem;
        text-align: center;
        box-shadow: 0 2px 5px var(--shadow-color);
    }

    .container {
        display: flex;
        flex: 1;
        overflow: hidden;
    }

    .palette {
        width: 220px;
        background-color: var(--palette-bg);
        padding: 1rem;
        border-right: 1px solid var(--border-color);
        box-shadow: 2px 0 5px var(--shadow-color);
        overflow-y: auto;
        display: grid;
        grid-template-columns: 1fr 1fr;
        gap: 10px;
        align-content: start;
        height: fit-content;
        max-height: 100%;
    }

    .workspace {
        flex: 1;
        background-color: var(--workspace-bg);
        padding: 2rem;
        overflow: auto;
        position: relative;
    }

    .palette-item {
        display: flex;
        flex-direction: column;
        align-items: center;
        justify-content: center;
        padding: 0.75rem 0.5rem;
        border: 1px solid var(--border-color);
        border-radius: 8px;
        cursor: pointer;
        transition: all 0.2s ease;
        user-select: none;
        background-color: var(--item-bg);
        font-size: 12px;
        color: var(--text-color);
        touch-action: none;
    }

    .palette-item:hover {
        transform: translateY(-2px);
        box-shadow: 0 3px 5px var(--shadow-color);
        background-color: var(--item-hover-bg);
    }

    .palette-item.selected {
        background-color: var(--selected-bg);
        border: 1px solid var(--selected-border);
        box-shadow: 0 0 0 2px var(--selected-border), 0 0 12px rgba(139, 92, 246, 0.5);
        position: relative;
        transform: scale(1.05);
    }

    .palette-item svg {
        margin-bottom: 8px;
        width: 36px;
        height: 36px;
    }

    .log-panel {
        padding: 1rem;
        background-color: #2c3e50;
        color: #ecf0f1;
        max-height: 150px;
        overflow-y: auto;
        font-family: monospace;
    }

    .log-entry {
        margin-bottom: 5px;
        border-bottom: 1px solid #1e293b;
        padding-bottom: 5px;
    }

    .log-entry:last-child {
        border-bottom: none;
    }

    /* Element previews in workspace */
    .element-preview {
        position: absolute;
        border: 2px dashed var(--primary-color);
        padding: 5px;
        border-radius: 4px;
        background-color: rgba(52, 152, 219, 0.1);
        min-width: 100px;
        min-height: 30px;
        display: flex;
        align-items: center;
        justify-content: center;
        font-size: 12px;
        color: #666;
    }

    /* Dragging Element Preview */
    .dragging-element {
        position: absolute;
        pointer-events: none;
        opacity: 0.8;
        z-index: 1000;
        transform: translate(-50%, -50%);
        background-color: var(--selected-bg);
        border: 1px solid var(--selected-border);
        padding: 5px 10px;
        border-radius: 4px;
        color: white;
        font-size: 12px;
    }
</style> 
              """


def _pretty(node):
    if node is None:
        return 'None'
    if hasattr(node, 'tagName'):
        return f'{node.tagName.lower()}#{node.id}.{node.className}[{node.innerHTML.strip()[:20]}…]'
    return str(node)



from __future__ import annotations

import logging

import js
from js import Array, document

from wwwpy.common.designer.html_locator import Node
from wwwpy.common.designer.locator_lib import Locator, Origin
from wwwpy.remote.component import get_component
from wwwpy.remote.jslib import is_instance_of

logger = logging.getLogger(__name__)


def locator_from(element: js.Element) -> Locator | None:
    """Returns an instance of ElementPath that describes the path to the element"""

    # Start building the path
    path = []
    current = element

    while current:
        # Check if we've reached a component
        component = get_component(current)
        if component:
            clazz = component.__class__
            # We found the component, return the path
            return Locator(clazz.__module__, clazz.__name__, path, Origin.live)

        # Reached document body without finding a component
        if current == document.body:
            return None

        # Handle shadow root - move to host element
        if is_instance_of(current, js.ShadowRoot):
            current = current.host
            continue

        # Get the parent to determine next step
        parent = current.parentNode
        if not parent:
            return None

        # Add the current node to the path
        add_node_to_path(current, parent, path)

        if hasattr(current, "assignedSlot") and current.assignedSlot:
            add_node_to_path(parent, parent.parentNode, path)
            current = parent.parentNode
            continue

        # Move up to parent for next iteration
        current = parent

    return None


def add_node_to_path(node, parent, path):
    """Helper to add a node to the path with its attributes and position"""
    child_index = Array.prototype.indexOf.call(parent.children, node) if parent else -1
    attributes = {attr.name: attr.value for attr in node.attributes}
    path.insert(0, Node(node.tagName.lower(), child_index, attributes))



import sys
import weakref
from typing import Callable


class _Obj:
    pass


def on_exit(callback: Callable[[], None]) -> None:
    frame = sys._getframe(1)
    obj = _Obj()
    frame.f_locals[obj] = obj
    weakref.finalize(obj, callback)



from __future__ import annotations

import logging
from typing import TypeVar, Type

import js

from wwwpy.common.designer.ui.rect_readonly import rect_xy_center
from wwwpy.remote.component import get_component, Component

logger = logging.getLogger(__name__)
ComponentType = TypeVar('ComponentType', bound=Component)


def global_instance(cls: Type[ComponentType], id_name: str | None = None) -> ComponentType:
    """
    Creates or retrieves a global instance of the given component class.

    Args:
        cls: The component class to instantiate
        id_name: Optional ID for the element, defaults to 'id_' + tag_name

    Returns:
        An instance of the specific component class
    """
    if id_name is None:
        id_name = 'id_' + cls.component_metadata.tag_name
    tag_name = cls.component_metadata.tag_name
    element = single_tag_instance(tag_name, id_name)
    comp = get_component(element)
    assert isinstance(comp, cls), f'Expected {cls}, tag_name={tag_name}, got {comp}, from element {element}'
    return comp


def single_tag_instance(tag_name: str, global_id: str, where=js.document.body) -> js.HTMLElement:
    """Get or create a single tag instance with the given ID."""
    ele = js.document.getElementById(global_id)
    if ele is None:
        ele = js.document.createElement(tag_name)
        ele.id = global_id
        where.appendChild(ele)
    return ele


def ensure_tag_instance(tag_name: str, global_id: str, where=js.document.body) -> js.HTMLElement:
    """Get or create a single tag instance with the given ID."""
    ele = js.document.getElementById(global_id)
    if ele is None:
        logger.debug(f'Creating new element {tag_name} with id {global_id}')
        ele = js.document.createElement(tag_name)
        ele.id = global_id
        where.appendChild(ele)
    return ele


def element_xy_center(element: js.Element) -> tuple[float, float]:
    """Get the center x and y coordinates of the given element relative to the page."""
    rect = element.getBoundingClientRect()
    return rect_xy_center(rect)



from __future__ import annotations

import logging
import time
from datetime import timedelta

logger = logging.getLogger(__name__)


class TimeLogger:
    def __init__(self, name):
        self.name = name
        self.start = time.perf_counter()

    def debug(self, message=''):
        logger.debug(self.message(message), stacklevel=2)

    def message(self, msg='') -> str:
        delta = self.time_spent()
        if msg:
            msg = f' - {msg}'
        return f'{self.name}{msg} time={delta}'

    def time_spent(self) -> timedelta:
        end = time.perf_counter()
        return timedelta(seconds=end - self.start)



from __future__ import annotations

import dataclasses
# todo maybe rename to locatorlib.py
import logging
from dataclasses import dataclass
from enum import Enum

from wwwpy.common.designer.html_locator import NodePath, data_name

logger = logging.getLogger(__name__)


class Origin(str, Enum):
    source = 'source'
    """The original html source in the component"""
    live = 'live'
    """The live html in the browser"""


@dataclass()
class Locator:
    """Contains the path to an element relative to a Component.
    This is intended to be serialized"""

    class_module: str
    """The module name of the Component."""
    class_name: str
    """The class name of the Component."""
    path: NodePath
    """The path from the Component (excluded) to the element."""

    origin: Origin

    @property
    def class_full_name(self) -> str:
        return f'{self.class_module}.{self.class_name}'

    @property
    def tag_name(self) -> str:
        """The tag name of the element in lowercase."""
        if len(self.path) == 0:
            return ''
        return self.path[-1].tag_name

    @property
    def data_name(self) -> str | None:
        """The name of the element in the data dictionary."""
        return data_name(self.path)

    def valid(self) -> bool:
        from wwwpy.common.designer import code_strings as cs, html_locator as hl
        html = cs.html_from(self.class_module, self.class_name)
        if not html:
            logger.warning(f'Cannot find html for {self.class_module}.{self.class_name}')
            return False
        span = hl.locate_span(html, self.path)
        if not span:
            logger.warning(f'Cannot locate span for {self.path} in html=`{html}`')
        return span is not None

    def rebase_to_origin(self) -> Locator | None:
        return rebase_to_origin(self)

    def match_component_type(self, component_type) -> bool:
        module = component_type.__module__
        name = component_type.__name__
        return module == self.class_module and name == self.class_name


def rebase_to_origin(locator: Locator) -> Locator | None:
    """
    This rebase from Origin.live to Origin.source
    It returns None if the locator is not valid.
    If the locator is already in Origin.source, it returns the locator unchanged.
    """
    if not locator:
        return None
    if locator.origin == Origin.source:
        return locator

    locator = dataclasses.replace(locator, origin=Origin.source)  # change origin to source immediately

    # todo probably this should go to some other file to avoid circular import issues
    from wwwpy.common.designer import code_strings, html_locator
    html = code_strings.html_from(locator.class_module, locator.class_name)
    if html is None:
        return None

    cst_node = html_locator.locate_node(html, locator.path)
    if cst_node is None:
        return locator

    node_path = html_locator.node_path_from_leaf(cst_node)
    result = dataclasses.replace(locator, path=node_path)
    return result



# Create a translation table for the specified characters
escape_table = str.maketrans({
    '\r': '\\r',  # Carriage return -> \r
    '\n': '\\n',  # Newline (line feed) -> \n
    '\t': '\\t',  # Tab -> \t
    '\\': '\\\\'  # Backslash -> \\
})


def escape_string(s: str) -> str:
    return s.translate(escape_table)


def unescape_string(s: str) -> str:
    escaped_bytes = s.encode('ascii', 'backslashreplace')
    return escaped_bytes.decode('unicode_escape')



from __future__ import annotations

from typing import TypeVar, Generic, Callable

T = TypeVar('T')


class TypeListeners(Generic[T], list[Callable[[T], None]]):
    def __init__(self, event_type: type[T] | None) -> None:
        super().__init__()
        self.event_type = event_type

    def add(self, handler: Callable[[T], None]) -> None:
        self.append(handler)

    def remove(self, handler: Callable[[T], None]) -> None:
        super().remove(handler)

    def notify(self, event: T) -> None:
        if self.event_type and not isinstance(event, self.event_type):
            raise TypeError(f'Handler expects {self.event_type}')
        for h in list(self):
            h(event)


class DictListeners:
    def __init__(self):
        self._listeners: dict[type, TypeListeners] = {}
        self.catch_all = TypeListeners(None)

    def on(self, event_type: type[T]) -> TypeListeners[T]:
        lst = self._listeners.get(event_type)
        if lst is None:
            lst = TypeListeners(event_type)
            self._listeners[event_type] = lst
        return lst

    def notify(self, ev: T) -> None:
        listeners = self._listeners.get(type(ev), None)
        if listeners:
            listeners.notify(ev)
        self.catch_all.notify(ev)



import logging

import wwwpy.remote.component as wpc
import wwwpy.remote.designer.ui.accordion_components  # noqa
import wwwpy.remote.designer.ui.comp_structure  # noqa
from wwwpy.common.designer.canvas_selection import CanvasSelection, CanvasSelectionChangeEvent
from wwwpy.common.designer.element_library import element_library
from wwwpy.common.injectorlib import injector
from wwwpy.remote.designer.ui import pushable_sidebar, palette
from wwwpy.remote.designer.ui.intent_add_element import AddElementIntent
from wwwpy.remote.designer.ui.intent_select_element import SelectElementIntent
from wwwpy.remote.designer.ui.property_editor import PropertyEditor
from wwwpy.remote.designer.ui.system_tools.system_tools_component import SystemToolsComponent

logger = logging.getLogger(__name__)


class NewToolbox(wpc.Component, tag_name='wwwpy-new-toolbox'):
    _sidebar: pushable_sidebar.PushableSidebar = wpc.element()
    _palette: palette.PaletteComponent = wpc.element()
    _property_editor: PropertyEditor = wpc.element()
    _system_tools: SystemToolsComponent = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """
<pushable-sidebar data-name="_sidebar" position="right" width="300px">
    <wwwpy-accordion-section expanded>
        <div slot="header">Add Components</div>
        <wwwpy-palette data-name="_palette" style="height: 300px; display: flex"></wwwpy-palette>
    </wwwpy-accordion-section>
    <wwwpy-accordion-section expanded>
        <div slot="header">Structure</div>
        <wwwpy-comp-structure style="height: 250px; display: flex; overflow: scroll"></wwwpy-comp-structure>
    </wwwpy-accordion-section>
    <wwwpy-accordion-section expanded>
        <div slot="header">Attributes/Events</div>
        <wwwpy-property-editor data-name="_property_editor" style="height: 250px; display: flex; overflow: scroll"></wwwpy-property-editor>
    </wwwpy-accordion-section>
    <wwwpy-accordion-section expanded>
        <div slot="header">System Tools</div>
        <wwwpy-system-tools data-name="_system_tools"></wwwpy-system-tools>
    </wwwpy-accordion-section>
</pushable-sidebar>
        """
        self._setup_palette()
        self._setup_property_editor()

    def _setup_palette(self):
        self._palette.add_intent(SelectElementIntent())
        for ei in element_library().elements:
            if ei.tag_name.startswith('sl-'):
                continue
            self._palette.add_intent(AddElementIntent(ei.tag_name, element_def=ei))

    def _setup_property_editor(self):
        def _csce(event: CanvasSelectionChangeEvent):
            logger.debug(f'canvas selection changed: {event}')
            self._property_editor.selected_element_path = event.new

        canvas_selection = injector.get(CanvasSelection)
        canvas_selection.on_change.add(_csce)

    @property
    def visible(self) -> bool:
        return self._sidebar.state != 'hidden'

    @visible.setter
    def visible(self, value: bool):
        self._sidebar.state = 'expanded' if value else 'hidden'



import js


def setup():
    js.document.head.insertAdjacentHTML('beforeend', simple_dark_theme_header)


# language=html
simple_dark_theme_header = """
<link rel="icon" href="data:,">
<meta name="viewport" content="width=device-width, initial-scale=1.0">
<style>
    :root {
        color-scheme: dark;
    }
    
    body {
        background: #121212;
        color: #e0e0e0;
        margin: 1rem;
        font: 16px sans-serif;
        line-height: 1.5;
    }

    a {
        color: #bb86fc;
        text-decoration: none;
    }

    a:hover {
        text-decoration: underline;
    }

    button, 
    input[type="button"], 
    input[type="submit"] {
        padding: 0.5rem 1rem;
        border: none;
        border-radius: 4px;
        font-size: 1rem;
        cursor: pointer;
        transition: background-color 0.2s;
        min-height: 40px;
    }

    button:hover, 
    input[type="button"]:hover, 
    input[type="submit"]:hover {
        background-color: #6200ee;
    }
    
    input[type="text"], 
    input[type="email"], 
    input[type="password"],
    input[type="number"],
    textarea {
        background-color: #1e1e1e;
        color: #ffffff;
        border: 1px solid #333;
        border-radius: 4px;
        padding: 0.5rem;
        font-size: 1rem;
        min-height: 40px;
        box-sizing: border-box;
        width: 100%;
    }

    select {
        border: 1px solid #333;
        border-radius: 4px;
        padding: 0.5rem;
        font-size: 1rem;
        min-height: 40px;
        width: 100%;
        appearance: none;
        background-image: url("data:image/svg+xml;utf8,<svg fill='white' height='24' width='24' xmlns='http://www.w3.org/2000/svg'><path d='M7 10l5 5 5-5z'/></svg>");
        background-repeat: no-repeat;
        background-position: right 0.5rem center;
        padding-right: 2rem;
    }

    input:focus, 
    select:focus, 
    textarea:focus {
        outline: none;
        border-color: #bb86fc;
    }

    /* Basic form layout */
    form {
        display: flex;
        flex-direction: column;
        gap: 1rem;
        max-width: 500px;
    }

    label {
        font-weight: bold;
        margin-bottom: 0.25rem;
        display: block;
    }

    /* Checkbox and radio styling */
    input[type="checkbox"],
    input[type="radio"] {
        accent-color: #bb86fc;
        margin-right: 0.5rem;
    }

    /* Table styling */
    table {
        border-collapse: collapse;
        width: 100%;
        margin: 1rem 0;
    }

    th, td {
        padding: 0.75rem;
        text-align: left;
        border-bottom: 1px solid #333;
    }

    th {
        background-color: #1e1e1e;
    }

    tr:hover {
        background-color: #1e1e1e;
    }
</style>
"""



from __future__ import annotations

import logging
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Iterator

from wwwpy.common.designer import code_info
from wwwpy.common.designer.code_strings import html_from_source
from wwwpy.common.designer.element_library import ElementDefBase
from wwwpy.common.designer.html_locator import node_path_from_leaf
from wwwpy.common.designer.html_parser import CstTree, html_to_tree, CstNode
from wwwpy.common.designer.locator_lib import Locator, Origin

logger = logging.getLogger(__name__)


class LocatorNode:
    # locator: Locator
    # children: list[LocatorNode]
    # cst_node: CstNode
    _cst_children: list[CstNode]

    def __init__(self, parent: CompInfo, cst: CstTree | CstNode):
        self._parent = parent
        if isinstance(cst, CstTree):
            self._cst_children = list(cst)
        elif isinstance(cst, CstNode):
            self._cst_children = list(cst.children)
            ep = node_path_from_leaf(cst)
            self.locator = Locator(self._parent.class_package, self._parent.class_name, ep, Origin.source)
            self.cst_node = cst
        else:
            raise TypeError(f'Expected CstTree or CstNode, got {type(cst)}')

    @cached_property
    def children(self) -> list[LocatorNode]:
        """The children of this node in the CST tree."""
        return [LocatorNode(self._parent, c) for c in self._cst_children]


# class LocatorTree(UserList[LocatorNode]):
#     parent: CompInfo
#     cst_tree: CstTree
#
#     def __init__(self, parent: CompInfo, cst: CstTree):
#         nodes = (LocatorNode())
#         super().__init__()


# todo extend CompInfo such as it provides the means for CompStructure component to show
#  the tree structure of the component's HTML.
#  think about a couple of design patterns to achieve this.
@dataclass
class CompInfo:
    """If the html is not found, no instance of this class is created."""
    class_package: str  # todo rename to class_module
    class_name: str
    tag_name: str
    path: Path
    cst_tree: CstTree

    @cached_property
    def locator_root(self) -> LocatorNode:
        """The CST tree of the component's HTML.
        This is lazy initialized"""
        return LocatorNode(self, self.cst_tree)

    @property
    def class_full_name(self) -> str:
        return f'{self.class_package}.{self.class_name}'


@dataclass
class ComponentDef(ElementDefBase):
    comp_info: CompInfo


def iter_comp_info_folder(folder: Path, package: str) -> Iterator[CompInfo]:
    """Iterate over all components in the folder."""
    logger.debug(f'Iterating over components in {folder}')
    for path in sorted(folder.glob('*.py'), key=lambda p: p.stem):
        yield from iter_comp_info(path, package + '.' + path.stem)


def iter_comp_info(path: Path, package: str) -> Iterator[CompInfo]:
    """Return a generator that yields CompInfo for each Component in the given python file."""
    logger.debug(f'Iterating over components in {path}')
    source_code = path.read_text()
    ci = code_info.info(source_code)
    return (c for c in (_to_comp_info(source_code, path, cl, package) for cl in ci.classes) if c is not None)


def _to_comp_info(source_code: str, path: Path, cl: code_info.ClassInfo, package: str) -> CompInfo | None:
    class_name = cl.name
    html = html_from_source(source_code, class_name)
    if html is None:
        logger.warning(f'Cannot find html for {class_name} in {path}')
        return None

    cst_tree = html_to_tree(html)

    return CompInfo(package, class_name, cl.tag_name, path, cst_tree)



from __future__ import annotations

import logging
from dataclasses import dataclass
from enum import Enum
from functools import cached_property

import js

from wwwpy.remote.designer.ui.locator_event import LocatorEvent

logger = logging.getLogger(__name__)


# todo we could remove IntentEvent and use directly js.PointerEvent
#  or we could use LocatorEvent. IntentManager could materialize LocatorEvent somehow (using injector)
@dataclass
class IntentEvent:
    js_event: js.PointerEvent

    @cached_property
    def deep_target(self) -> js.HTMLElement | None:
        from wwwpy.remote.jslib import get_deepest_element
        return get_deepest_element(self.js_event.clientX, self.js_event.clientY)


def js_event_to_locator_event_default(js_event: js.PointerEvent) -> LocatorEvent | None:
    """Convert a JavaScript PointerEvent to a LocatorEvent."""
    from wwwpy.remote.designer.ui.design_aware import ep_is_selectable_le
    locator_event = LocatorEvent.from_pointer_event(js_event)

    from wwwpy.remote.designer.ui.design_aware import ep_locator_event_transformer
    le_transformed = ep_locator_event_transformer(locator_event)

    if le_transformed is not None:
        return le_transformed

    if not ep_is_selectable_le(locator_event):
        return None

    return locator_event


@dataclass
class Intent:
    label: str
    """Label to be displayed in the palette item."""

    selected: bool = False
    """True if the item is selected, False otherwise."""

    def on_selected(self):
        ...

    def on_hover(self, _: LocatorEvent):
        ...

    def on_submit(self, _: LocatorEvent) -> bool:
        """If return True, the intent is deselected, if False, it is not."""

    def on_deselected(self):
        ...

    def on_hover_js(self, js_event: js.PointerEvent):
        """Handle hover event from JavaScript."""
        event = js_event_to_locator_event_default(js_event)
        if event:
            self.on_hover(event)

    def on_submit_js(self, js_event: js.PointerEvent) -> bool:
        """Handle submit event from JavaScript."""
        event = js_event_to_locator_event_default(js_event)
        if event:
            return self.on_submit(event)
        return False


@dataclass
class IntentChangedEvent:
    old: Intent | None
    new: Intent | None


class Support(str, Enum):
    CONTAINER = 'CONTAINER'



from __future__ import annotations

import logging
from typing import Callable

import libcst as cst

from wwwpy.common import modlib

logger = logging.getLogger(__name__)


def html_string_edit(source_code: str, class_name: str, html_manipulator: Callable[[str], str]) -> str:
    """This function modifies the HTML string in the source code using the provided manipulator function.
    It only modifies the HTML string within the specified class.
    Supports both triple single-quoted and triple double-quoted strings.
    """
    tree = cst.parse_module(source_code)
    transformer = _HTMLStringUpdater(class_name, html_manipulator)
    modified_tree = tree.visit(transformer)
    return modified_tree.code


def html_from(module: str, class_name: str) -> str | None:
    path = modlib._find_module_path(module)
    if not path:
        return None
    source_code = path.read_text()
    return html_from_source(source_code, class_name)


def html_from_source(source_code: str, class_name: str) -> str | None:
    """This function extracts the HTML string from the source code."""
    logger.debug(f'html_from_source {class_name} source_code=`{source_code}`')

    html_res = []

    def html_manipulator(html):
        html_res.append(html)
        return html

    html_string_edit(source_code, class_name, html_manipulator)  # improper use to capture html
    if not html_res:
        return None
    return html_res[0]


class _HTMLStringUpdater(cst.CSTTransformer):

    def __init__(self, class_name: str, html_manipulator: Callable[[str], str]):
        super().__init__()
        self.class_name = class_name
        self.html_manipulator = html_manipulator
        self.current_class = None
        self.current_method = None
        self.in_assignment = False
        self.is_innerHTML_target = False
        self.found_assignment = False

    def visit_ClassDef(self, node: cst.ClassDef) -> None:
        self.current_class = node.name.value

    def leave_ClassDef(self, original_node: cst.ClassDef, updated_node: cst.ClassDef) -> cst.CSTNode:
        self.current_class = None
        return updated_node

    def visit_FunctionDef(self, node: cst.FunctionDef) -> None:
        self.current_method = node.name.value
        self.found_assignment = False

    def leave_FunctionDef(self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef) -> cst.CSTNode:
        self.current_method = None
        return updated_node

    def visit_Assign(self, node: cst.Assign) -> None:
        self.in_assignment = True
        # Check if the target contains 'innerHTML'
        for target in node.targets:
            if isinstance(target.target, cst.Attribute):
                attr = target.target
                while isinstance(attr, cst.Attribute):
                    if 'innerHTML' in attr.attr.value:
                        self.is_innerHTML_target = True
                        break
                    attr = attr.value

    def leave_Assign(self, original_node: cst.Assign, updated_node: cst.Assign) -> cst.CSTNode:
        self.in_assignment = False
        self.is_innerHTML_target = False
        return updated_node

    def leave_SimpleString(self, original_node: cst.SimpleString, updated_node: cst.SimpleString) -> cst.CSTNode:
        if (self.current_class == self.class_name and
                self.current_method == "init_component" and
                self.in_assignment and
                self.is_innerHTML_target and
                not self.found_assignment):
            
            if (original_node.value.startswith('"""') and original_node.value.endswith('"""')) or \
                    (original_node.value.startswith("'''") and original_node.value.endswith("'''")):
                quote_type = original_node.value[:3]
                original_html = original_node.value[3:-3]
                modified_html = self.html_manipulator(original_html)
                self.found_assignment = True
                return updated_node.with_changes(value=f'{quote_type}{modified_html}{quote_type}')
        return updated_node



from __future__ import annotations

import logging
from pathlib import Path

import js

import wwwpy.remote.component as wpc
from wwwpy.common.designer.ui.svg import add_rounded_background2
from wwwpy.remote import dict_to_js

logger = logging.getLogger(__name__)

_BLUE = '#3574F0'
_GRAY = '#3C3E41'
_BGRD = '#2B2D30'


class SvgIcon(wpc.Component, tag_name='wwwpy-svg-icon'):
    filename: str = wpc.attribute()
    _style: js.HTMLStyleElement = wpc.element()
    _div: js.HTMLDivElement = wpc.element()
    _active: bool
    border: int = 5

    @classmethod
    def from_file(cls, file: Path) -> SvgIcon:
        r = cls()
        r.load_svg_str(file.read_text())
        r.element.setAttribute('title', file.name)
        return r

    def init_component(self):
        self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.element.shadowRoot.innerHTML = """
<style data-name="_style"></style>
<div data-name="_div"></div>
        """

        self.active = False

    def load_svg_str(self, svg: str):
        self._div.innerHTML = add_rounded_background2(svg, 'var(--svg-primary-color)')

    @property
    def active(self) -> bool:
        # todo probably rename to 'selected' and add new property 'disabled'
        #  if disabled, change the colors accordingly; if not disabled honor the colors from the active property
        return self._active

    @active.setter
    def active(self, value: bool):
        self._active = value
        root, hover = (_BLUE, '') if value else (_BGRD, _GRAY)
        self._set_style(root, hover)

    def _set_style(self, color: str, hover_color: str):
        # language=html
        hover_style = ":host(:hover) { --svg-primary-color: %s; }" % (hover_color,) if hover_color else ''
        s = ('svg { display: block }\n' +
             ':host { border: %spx solid transparent }\n' % self.border +
             ':host { --svg-primary-color: %s; }\n' % color + hover_style)
        logger.debug(f'set_style: `{s}`')
        self._style.innerHTML = s

    def _div__click(self, event):
        self.active = not self.active



from wwwpy.remote.designer.ui import palette, pushable_sidebar, floater_action_band


def register_bindings():
    palette.extension_point_register()
    pushable_sidebar.register_extension_point()
    floater_action_band.extension_point_register()



from typing import Protocol


class RectReadOnly(Protocol):
    """This is intended to be used in common to handle js.DOMRectReadOnly"""
    x: float
    y: float
    width: float
    height: float
    top: float
    right: float
    bottom: float
    left: float

    def toJSON(self) -> object: ...


def rect_xy_center(rect: RectReadOnly) -> tuple[float, float]:
    """Get the center x and y coordinates of the given rect."""
    x = rect.x + rect.width / 2
    y = rect.y + rect.height / 2
    return x, y


def rect_to_py(rect: RectReadOnly) -> RectReadOnly:
    from . import rect_readonly_py
    return rect_readonly_py.RectReadOnlyPy(rect)



from __future__ import annotations

import asyncio
import logging

import js

from wwwpy.common.injectorlib import injector
from wwwpy.remote.designer.ui.element_selector import ElementSelector

logger = logging.getLogger(__name__)


async def set_active(dev_mode: bool):
    js.window._wwwpy_dev_mode = dev_mode
    if dev_mode:
        await _activate()


def is_active() -> bool:
    try:
        return js.window._wwwpy_dev_mode
    except AttributeError:
        return False


async def _activate():
    from wwwpy.remote import micropip_install
    from wwwpy.common import designer

    for package in designer.pypi_packages:
        await micropip_install(package)

    from wwwpy.remote.designer import helpers
    js.window.onerror = helpers._on_error
    js.window.onunhandledrejection = helpers._on_unhandledrejection
    from wwwpy.remote.designer import log_redirect
    log_redirect.redirect_logging()

    loop = asyncio.get_event_loop()
    loop.set_exception_handler(_global_exception_handler)

    # dependency injection
    injector._clear()

    from wwwpy.common.eventbus import EventBus
    injector.bind(EventBus())  # todo, maybe use a named binding so we use our own event bus
    from wwwpy.remote.designer.ui.intent_manager import IntentManager
    intent_manager = IntentManager()
    intent_manager.install()
    injector.bind(intent_manager)
    injector.bind(ElementSelector())
    from wwwpy.common.designer.canvas_selection import CanvasSelection
    injector.bind(CanvasSelection())
    from wwwpy.remote.designer import di_remote
    di_remote.register_bindings()


def on_after_remote_main():
    if not is_active():
        return
    from wwwpy.remote.designer.ui import dev_mode_component
    dev_mode_component.show()  # todo it looks like it can take ~ 1s or more; investigate,
    # maybe micropip installing rope has to do with it


def _global_exception_handler(loop, context):
    # The context parameter contains details about the exception
    if 'message' in context:
        logger.info(f"Global handler caught: {context['message']}")
    if 'exception' not in context:
        logger.error(f"Global handler caught: {context}")
        return
    exception = context.get('exception')
    if exception:
        logger.info(f"Exception type: {type(exception)}, Args: {exception.args}")
        logger.exception(exception)



import logging

from wwwpy.common.designer.html_edit import Position

logger = logging.getLogger(__name__)


def position_for(width: float, height: float, x: float, y: float) -> Position:
    ratio = 0.35
    iw = width * ratio
    ih = height * ratio
    x0 = (width - iw) / 2
    y0 = (height - ih) / 2
    inside_inner = x0 <= x <= x0 + iw and y0 <= y <= y0 + ih
    if inside_inner:
        ix = x - x0
        iy = y - y0
        tl = _is_top_left(iw, ih, ix, iy)
        res = Position.afterbegin if tl else Position.beforeend
    else:
        tl = _is_top_left(width, height, x, y)
        res = Position.beforebegin if tl else Position.afterend
    logger.debug(f'position_for: x={x}, y={y}, iw={iw}, ih={ih}, x0={x0}, y0={y0}, res={res}')
    return res


def _is_top_left(w: float, h: float, x: float, y: float) -> bool:
    return y <= -h / w * x + h


def svg_indicator_for(width: float, height: float, position: Position) -> str:
    inactive_color = 'gray'
    active_color = 'green'
    iw = width * 25 / 70
    ih = height * 10 / 30
    x = (width - iw) / 2
    y = (height - ih) / 2
    out_tl = position == Position.beforebegin  # outer top left
    out_br = position == Position.afterend  # outer bottom right
    inn_tl = position == Position.afterbegin  # inner top left
    inn_br = position == Position.beforeend  # inner bottom right

    out_tl_color = active_color if out_tl else inactive_color
    out_br_color = active_color if out_br else inactive_color
    out_tl_width = 6 if out_tl else 2
    out_br_width = 6 if out_br else 2

    inn_tl_color = active_color if inn_tl else inactive_color
    inn_br_color = active_color if inn_br else inactive_color
    inn_tl_width = 3 if inn_tl else 1
    inn_br_width = 3 if inn_br else 1

    # language=html
    svg = '''<svg width="%(w)s" height="%(h)s" viewBox="0 0 %(w)s %(h)s" xmlns="http://www.w3.org/2000/svg">
  <g stroke="%(out_tl_color)s" stroke-width="%(out_tl_width)s">
    <line x1="0" y1="0" x2="%(w)s" y2="0"/>
    <line x1="0" y1="0" x2="0" y2="%(h)s"/>
  </g>
  <g stroke="%(out_br_color)s" stroke-width="%(out_br_width)s">
    <line x1="0" y1="%(h)s" x2="%(w)s" y2="%(h)s"/>
    <line x1="%(w)s" y1="0" x2="%(w)s" y2="%(h)s"/>
  </g>
  
  <g stroke="%(inn_tl_color)s" stroke-width="%(inn_tl_width)s" transform="translate(%(x)s, %(y)s)">
    <line x1="0" y1="0" x2="%(iw)s" y2="0"/>
    <line x1="0" y1="0" x2="0" y2="%(ih)s"/>
  </g>
  <g stroke="%(inn_br_color)s" stroke-width="%(inn_br_width)s" transform="translate(%(x)s, %(y)s)">
    <line x1="0" y1="%(ih)s" x2="%(iw)s" y2="%(ih)s"/>
    <line x1="%(iw)s" y1="0" x2="%(iw)s" y2="%(ih)s"/>
  </g>
  
</svg>'''
    return svg % dict(
        w=width, h=height,
        iw=iw, ih=ih,
        x=x, y=y,
        out_tl_color=out_tl_color, out_br_color=out_br_color,
        out_tl_width=out_tl_width, out_br_width=out_br_width,

        inn_tl_width=inn_tl_width, inn_tl_color=inn_tl_color,
        inn_br_width=inn_br_width, inn_br_color=inn_br_color,
    )


# language=html
_animate = """<animate attributeName="stroke" values="black;white;black" dur="1.5s" repeatCount="indefinite"/>"""

# language=html
_animate_transform = """
<animateTransform attributeName="transform" type="translate" values="3,3; 10,10; 3,3"
                dur="1s" repeatCount="indefinite" additive="sum"/>
                """
# language=html
_svg_at_begin = f"""
<svg [attrs] viewBox="0 0 100 100" xmlns="http://www.w3.org/2000/svg">
    <g>
        <!-- Main diagonal line (arrow body) - shorter -->
        <line x1="70" y1="70" x2="0" y2="0" stroke-width="6" stroke="black">{_animate}</line>
        <!-- First arrowhead segment -->
        <line x1="-3" y1="0" x2="47" y2="0" stroke-width="6" stroke="black">{_animate}</line>
        <!-- Second arrowhead segment -->
        <line x1="0" y1="-3" x2="0" y2="47" stroke-width="6" stroke="black">{_animate}</line>
        
        {_animate_transform}
    </g>
</svg>
"""
# language=html
_svg_at_end = f"""

<svg  [attrs] viewBox="0 0 100 100" xmlns="http://www.w3.org/2000/svg">
    <g>
        <!-- Main diagonal line (arrow body) - shorter -->
        <line x1="30" y1="30" x2="100" y2="100" stroke-width="6" stroke="black">{_animate}</line>
        <!-- First arrowhead segment (horizontal) -->
        <line x1="53" y1="100" x2="103" y2="100" stroke-width="6" stroke="black">{_animate}</line>
        <!-- Second arrowhead segment (vertical) -->
        <line x1="100" y1="53" x2="100" y2="103" stroke-width="6" stroke="black">{_animate}</line>
        <animateTransform
                attributeName="transform"
                type="translate"
                values="-3,-3; -10,-10; -3,-3"
                dur="1s"
                repeatCount="indefinite"
                additive="sum"/>
    </g>
</svg>
"""

# language=html
_svg_at_inside = f"""
<svg [attrs] viewBox="0 0 100 100" xmlns="http://www.w3.org/2000/svg">
todo
</svg>
"""



import logging

import js

import wwwpy.remote.component as wpc
from wwwpy.remote.jslib import waitAnimationFrame

logger = logging.getLogger(__name__)


class SystemVersions(wpc.Component, tag_name='wwwpy-system-versions'):
    _container: js.HTMLElement = wpc.element()
    _wwwpy: js.HTMLElement = wpc.element()
    _python_remote: js.HTMLElement = wpc.element()
    _python_server: js.HTMLElement = wpc.element()
    _pyodide: js.HTMLElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """
        <pre data-name="_container">
    <span data-name="_wwwpy">wwwpy==...</span>
    <span data-name="_pyodide">pyodide==...</span>
    
    Python versions:
    <span data-name="_python_remote">pyton remote==...</span>
    <span data-name="_python_server">pyton server==...</span>
</pre>"""

    async def after_init_component(self):
        await waitAnimationFrame()

        import platform
        pyver = platform.python_version()

        try:
            import wwwpy
            wp_ver = wwwpy.__version__
        except Exception as e:
            wp_ver = e.message

        try:
            import js
            pyodide_ver = js.window.pyodide.version
        except Exception as e:
            pyodide_ver = e.message

        self._wwwpy.innerText = f'wwwpy==' + wp_ver
        self._pyodide.innerText = 'pyodide==' + pyodide_ver
        self._python_remote.innerText = 'remote==' + pyver + ' (this is tied to Pyodide version)'

        from wwwpy.server.designer import rpc
        server_version = await rpc.server_python_version_string()
        self._python_server.innerText = 'server==' + server_version



from js import fetch, console

from wwwpy.common.fetch_debug import fetch_debug

# logger = logging.getLogger(__name__)
logger = console  # if log redirect is active, it will create an infinite loop because


async def async_fetch_str(url: str, method: str = 'GET', data: str = '') -> str:
    logger.debug(__name__ + ' ' + fetch_debug(url, method, data))
    response = await fetch(url, method=method, body=data)
    text = await response.text()
    return text



from __future__ import annotations

import logging

import js
from pyodide.ffi import create_proxy

import wwwpy.remote.component as wpc
from wwwpy.remote import dict_to_js
from wwwpy.remote.component import get_component

logger = logging.getLogger(__name__)


class AccordionContainer(wpc.Component, tag_name='wwwpy-accordion-container'):
    def init_component(self):
        self.element.attachShadow(dict_to_js({'mode': 'open'}))
        self.element.shadowRoot.innerHTML = '<slot></slot>'

    @property
    def sections(self) -> list[AccordionSection]:
        return [
            get_component(child) for child in self.element.children
            if child.tagName.lower() == 'wwwpy-accordion-section'
        ]

    def collapse_all(self):
        for section in self.sections:
            section.expanded = False

    def expand_all(self):
        for section in self.sections:
            section.expanded = True


class AccordionSection(wpc.Component, tag_name='wwwpy-accordion-section'):
    _header_container: js.HTMLElement = wpc.element()
    _panel_container: js.HTMLElement = wpc.element()
    _expanded: bool
    _expanded_attr: str = wpc.attribute('expanded')

    def init_component(self):
        self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.element.shadowRoot.innerHTML = """
<slot name="header" data-name="_header_container">Header</slot>
<div data-name="_panel_container" style="display: grid">
    <div style="overflow: hidden">
        <slot>Panel</slot>
    </div>
</div>
"""
        self._expanded = None  # type: ignore
        self._copy_attribute_to_property()
        self.transition = True

    def _copy_attribute_to_property(self):
        self.expanded = True if self.element.hasAttribute('expanded') else False

    def attributeChangedCallback(self, name: str, oldValue: str | None, newValue: str | None):
        if name == 'expanded':
            self._copy_attribute_to_property()

    def _header_container__click(self, event):
        self.toggle(True)

    @property
    def transition(self):
        return self._panel_container.style.transition != ''

    @transition.setter
    def transition(self, value):
        self._panel_container.style.transition = '' if not value else 'grid-template-rows 300ms ease-in-out'

    @property
    def expanded(self):
        return self._expanded

    @expanded.setter
    def expanded(self, value):
        new_value = bool(value)
        if self._expanded == new_value:
            return
        self._expanded = new_value
        self._panel_container.style.gridTemplateRows = '1fr' if self._expanded else '0fr'
        self._expanded_attr = '' if new_value else None

    def toggle(self, emit_event=False):
        self.expanded = not self.expanded
        if not emit_event:
            return
        event_dict = dict_to_js({'bubbles': True, 'detail': {'section': self, 'expanded': self._expanded}})
        event = js.CustomEvent.new('accordion-toggle', event_dict)
        self.element.dispatchEvent(event)



from __future__ import annotations

import logging
from dataclasses import dataclass

from wwwpy.common.designer.locator_lib import Locator
from wwwpy.common.type_listener import TypeListeners

logger = logging.getLogger(__name__)


@dataclass
class CanvasSelectionChangeEvent:
    old: Locator | None
    new: Locator | None


class CanvasSelection:
    """At the time of writing, this class is listened by the old infrastructure, but
    it is not updated by it. Who writes it is the new infrastructure.
    """
    current_selection: Locator | None
    """It looks like the ElementPath should have Origin.live"""

    def __init__(self):
        self._current_selection = None
        self.on_change: TypeListeners[CanvasSelectionChangeEvent] = TypeListeners(CanvasSelectionChangeEvent)

    @property
    def current_selection(self) -> Locator | None:
        return self._current_selection

    @current_selection.setter
    def current_selection(self, value: Locator | None):
        old = self._current_selection
        self._current_selection = value
        event = CanvasSelectionChangeEvent(old, value)
        logger.debug(f'CanvasSelection: {old} -> {value}')
        self.on_change.notify(event)



from __future__ import annotations

import js

from wwwpy.common.designer.ui.rect_readonly import RectReadOnly
from wwwpy.remote import dict_to_js
from wwwpy.remote.designer.ui.floater import Floater


class WeirdSelectionIndicatorFloater(Floater, tag_name='weird-selection-indicator-floater'):

    def init_component(self):
        self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.element.shadowRoot.innerHTML = """
        <style>
            :host {
              position: fixed;
              pointer-events: none;
              border: 2px solid #4a90e2;
              z-index: 200000;
              display: none;
            } 
        </style>      
        """

    @property
    def transition(self) -> bool:
        return self.element.style.transition != 'none'

    @transition.setter
    def transition(self, value: bool):
        if value:
            self.element.style.transition = 'all 0.2s ease'
        else:
            self.element.style.transition = 'none'

    @property
    def visible(self) -> bool:
        return self.element.style.display == 'block'

    def hide(self):
        self.element.style.display = 'none'

    def set_reference_geometry(self, rect: RectReadOnly):
        bs = 2  # Adjust this value to match the border size in CSS

        r = js.DOMRect.new(rect.x - bs, rect.y - bs, rect.width, rect.height, )
        # r = rect

        self.element.style.display = 'block'
        self.element.style.top = f"{r.top}px"
        self.element.style.left = f"{r.left}px"
        self.element.style.width = f"{r.width}px"
        self.element.style.height = f"{r.height}px"

        sr = self.element.shadowRoot
        while sr.children.length > 1:
            sr.children[1].remove()

        x, y = compute_xy(rect.width, rect.height)
        fragment = js.document.createRange().createContextualFragment(
            create_svg(rect.width, rect.height, x, y, 'inner'))
        sr.appendChild(fragment)


def compute_xy(W, H):
    import math
    k = (1 - 1 / math.sqrt(3)) / 2
    return round(W * k, 9), round(H * k)


def create_svg(w, h, x, y, pulse_option) -> str:
    """Create SVG string based on parameters."""
    # Validate to ensure inner rectangle has positive dimensions
    if x * 2 >= w or y * 2 >= h:
        return f'<text x="10" y="30" fill="red">Error: Inset values too large for the given width/height</text>'

    # Calculate inner rectangle dimensions
    inner_width = w - 2 * x
    inner_height = h - 2 * y

    # Determine which parts should pulse
    top_class, left_class, bottom_class, right_class, inner_class = [''] * 5

    pulse = 'class="pulse" stroke="cyan"'
    if pulse_option == "top_left":
        top_class, left_class = pulse, pulse
    elif pulse_option == "bottom_right":
        bottom_class, right_class = pulse, pulse
    elif pulse_option == "inner":
        inner_class = pulse

    # Create SVG string with the CSS classes for animation
    return f'''<svg viewBox="0 0 {w} {h}" width="{w}" height="{h}" xmlns="http://www.w3.org/2000/svg">
            <!-- Outer rectangle segments -->
            <line x1="0" y1="0" x2="{w}" y2="0" stroke="blue" stroke-width="2" {top_class} />
            <line x1="0" y1="0" x2="0" y2="{h}" stroke="blue" stroke-width="2" {left_class} />
            <line x1="0" y1="{h}" x2="{w}" y2="{h}" stroke="blue" stroke-width="2" {bottom_class} />
            <line x1="{w}" y1="0" x2="{w}" y2="{h}" stroke="blue" stroke-width="2" {right_class} />

            <!-- Inner rectangle -->
            <rect x="{x}" y="{y}" width="{inner_width}" height="{inner_height}" fill="none" stroke="red" stroke-width="2" {inner_class} />

            <!-- Line connecting bottom-left corners -->
            <line x1="0" y1="{h}" x2="{x}" y2="{h - y}" stroke="green" stroke-width="2" />

            <!-- Line connecting top-right corners -->
            <line x1="{w}" y1="0" x2="{w - x}" y2="{y}" stroke="purple" stroke-width="2" />
        </svg>'''
